/* 
   elmo - ELectronic Mail Operator

   Copyright (C) 2003 rzyjontko

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  

   ----------------------------------------------------------------------

   This file contains an implementation of an address book.

*/
/****************************************************************************
 *    IMPLEMENTATION HEADERS
 ****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <ctype.h>

#include "address.h"
#include "raddress.h"
#include "error.h"
#include "file.h"
#include "xmalloc.h"
#include "ask.h"
#include "cmd.h"
#include "mail.h"
#include "select.h"
#include "folder.h"
#include "line.h"
#include "read.h"
#include "gettext.h"
#include "memblock.h"
#include "confread.h"
#include "abook.h"
#include "eprintf.h"
#include "str.h"
#include "sender.h"

/****************************************************************************
 *    IMPLEMENTATION PRIVATE DEFINITIONS / ENUMERATIONS / SIMPLE TYPEDEFS
 ****************************************************************************/

#define INITIAL_BLOCK_SIZE 2048
#define ABOOK_WIDTH 80

#define DFL_LOCATION() file_with_dir (getenv ("HOME"), ".addressbook")
#define NONEMPTY do { if (abook->count == 0) return; } while (0)
#define PREAMBLE do { if (abook_select == NULL) return; } while (0)
#define PREAMBLE_ADD do { if (add_select == NULL) return; } while (0)

#define DEFAULT_FORMAT "%s %020n %e"

/****************************************************************************
 *    IMPLEMENTATION PRIVATE CLASS PROTOTYPES / EXTERNAL CLASS REFERENCES
 ****************************************************************************/

/**
 * There is only one structure of this type.  It is used to collect data
 * about the address being read from the addressbook file.
 */
typedef struct new_item {
  address_t  *addr;
  char   **s_field;
  int      shift;
  unsigned mask;
} new_item_t;

/****************************************************************************
 *    IMPLEMENTATION PRIVATE STRUCTURES / UTILITY CLASSES
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION REQUIRED EXTERNAL REFERENCES (AVOID)
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE DATA
 ****************************************************************************/

/**
 * This is to decide if the addressbook should be flushed back to the file.
 */
static int abook_changed = 0;

/**
 * This is to decide if the addressbook should be sorted before showing
 * it to the user.
 */
static int abook_sorted = 0;

/**
 * This is the address book itself.  It's just a set of addresses grouped
 * in a single array.
 */
static raddress_t *abook = NULL;

/**
 * These are used when opening a window with addresses which should be added
 * to the address book.
 */
static select_t   *add_select   = NULL;
static raddress_t *add_raddress = NULL;

/**
 * Selection window with addresses from address book.
 */
static select_t   *abook_select = NULL;

/**
 * This is used when reading configuration.
 */
static new_item_t new_item = {NULL, NULL, 0, 0};

/**
 * All addressbook addresses (with their data) are stored in this memblock.
 */
static memblock_t *addr_block = NULL;

/****************************************************************************
 *    INTERFACE DATA
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE FUNCTION PROTOTYPES
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE FUNCTIONS
 ****************************************************************************/


static rstring_t *
make_items (void)
{
  address_t **addr;
  char       *fmt = ask_for_default ("abook_format", NULL);
  rstring_t  *result;

  if (abook == NULL)
    return NULL;

  result = rstring_create_size (abook->count + 1);

  if (fmt == NULL)
    fmt = DEFAULT_FORMAT;

  result->allocated_all = 1;  
  for (addr = abook->array; *addr; addr++){
    rstring_add (result, eprintf_addr (fmt, *addr));
  }
  return result;
}



static int
int_value (const char *s)
{
  if (strcmp (s, "M") == 0)
    return SEX_MALE;
  if (strcmp (s, "m") == 0)
    return SEX_MALE;
  if (strcmp (s, "male") == 0)
    return SEX_MALE;

  if (strcmp (s, "F") == 0)
    return SEX_FEMALE;
  if (strcmp (s, "f") == 0)
    return SEX_FEMALE;
  if (strcmp (s, "female") == 0)
    return SEX_FEMALE;

  if (strcmp (s, "Y") == 0)
    return YES;
  if (strcmp (s, "y") == 0)
    return YES;
  if (strcmp (s, "yes") == 0)
    return YES;
  if (strcmp (s, "YES") == 0)
    return YES;
  if (strcmp (s, "1") == 0)
    return YES;
  if (strcmp (s, "true") == 0)
    return YES;
  if (strcmp (s, "t") == 0)
    return YES;
  if (strcmp (s, "T") == 0)
    return YES;

  return NO;
}


static void
write_addr (address_t *addr, FILE *fp)
{
  fprintf (fp, "addr {\n");

  /**
   * it won't work if name has both kinds of quotes (single and double)
   */
  if (addr->name){
    if (addr->flags.bits.atomic_name)
      fprintf (fp, " name:\t\t%s\n", addr->name);
    else if (! addr->flags.bits.quotes_in_name)
      fprintf (fp, " name:\t\t\"%s\"\n", addr->name);
    else
      fprintf (fp, " name:\t\t'%s'\n", addr->name);
  }
  if (addr->email)
    fprintf (fp, " email:\t\t%s\n", addr->email);
  if (addr->initials)
    fprintf (fp, " initials:\t%s\n", addr->initials);
  if (addr->flags.bits.sex == SEX_MALE)
    fprintf (fp, " sex:\t\tM\n");
  else if (addr->flags.bits.sex == SEX_FEMALE)
    fprintf (fp, " sex:\t\tF\n");
  if (addr->flags.bits.official)
    fprintf (fp, " official:\tY\n");
  if (addr->flags.bits.foreign)
    fprintf (fp, " foreign:\tY\n");

  fprintf (fp, "}\n");
}


static int
cmp_addr (const void *a, const void *b)
{
  const address_t **aa = (const address_t **) a;
  const address_t **bb = (const address_t **) b;

  return address_cmp (*aa, *bb);
}


/****************************************************************************
 *    INTERFACE FUNCTIONS
 ****************************************************************************/

void
abook_init (void)
{

}



void
abook_free_resources (void)
{
  if (addr_block){
    memblock_destroy (addr_block);
    addr_block = NULL;
  }

  if (abook){
    raddress_destroy (abook);
    abook = NULL;
  }
}



void
abook_put_loaded (void)
{
  int i;
  
  if (abook == NULL)
    abook_load ();

  if (abook == NULL)
    return;

  for (i = 0; i < abook->count; i++)
    address_complete (abook->array[i]);
}



void
abook_load (void)
{
  char *dfl_location = DFL_LOCATION ();
  char *fname        = ask_for_default ("addressbook", NULL);

  if (fname == NULL)
    fname = dfl_location;

  abook_free_resources ();
  addr_block   = memblock_create (INITIAL_BLOCK_SIZE);
  abook        = raddress_create ();
  abook_sorted = 0;
  confread_read_file (fname);
  
  xfree (dfl_location);
}


void
abook_save (void)
{
  int     i;
  address_t *addr;
  char   *dfl_location;
  char   *fname;
  FILE   *fp;


  if (! abook_changed)
    return;

  dfl_location = DFL_LOCATION ();
  fname        = ask_for_default ("addressbook", NULL);

  if (fname == NULL)
    fname = dfl_location;

  fp = fopen (fname, "w");
  if (fp == NULL){
    error_ (errno, "%s", fname);
    xfree (dfl_location);
    return;
  }

  for (i = 0; i < abook->count; i++){
    addr = abook->array[i];
    write_addr (addr, fp);
  }

  fclose (fp);
  xfree (dfl_location);
}



void
abook_sort (void)
{
  if (abook == NULL)
    return;
  
  if (abook_sorted)
    return;

  /* do not sort addressbook when selection window is open */
  if (abook_select)
    return;
  
  qsort (abook->array, abook->count, sizeof (address_t *), cmp_addr);
  abook_sorted = 1;
}


rstring_t *
abook_strings (const char *prefix)
{
  int        i;
  char      *str;
  address_t    *addr;
  rstring_t *result = rstring_create_size (abook->count + 1);

  if (*prefix == '"')
    prefix++;
  
  for (i = 0; i < abook->count; i++){
    addr = abook->array[i];
    if (*addr->full == '"')
      str = addr->full + 1;
    else
      str = addr->full;
    if (strstr (str, prefix) == str)
      rstring_add (result, abook->array[i]->full);
  }
  
  return result;
}

/****************************************************************************
 *    ADDRESSBOOK WINDOW PRIVATE DATA
 ****************************************************************************/

/**
 * This string holds list of addresses that may be a value of To field.
 */
static str_t *to_str = NULL;

/****************************************************************************
 *    ADDRESSBOOK WINDOW PRIVATE FUNCTIONS
 ****************************************************************************/

static void
action_remove (int index)
{
  address_t *addr = abook->array[index];

  addr->flags.bits.abook = NO;
  
  abook_changed = 1;
  abook_sorted  = 0;
  raddress_remove (abook, index);
}



static void
action_change_sex (int index)
{
  address_t *addr = abook->array[index];

  if (addr->flags.bits.sex == SEX_MALE)
    addr->flags.bits.sex = SEX_FEMALE;
  else if (addr->flags.bits.sex == SEX_FEMALE)
    addr->flags.bits.sex = SEX_MALE;

  if (addr->flags.bits.sex != SEX_UNKNOWN)
    abook_changed = 1;
}


static void
action_change_official (int index)
{
  address_t *addr = abook->array[index];

  addr->flags.bits.official = ! addr->flags.bits.official;
  abook_changed             = 1;
}


static void
action_add_to_str (int index)
{
  address_t *addr = abook->array[index];
  
  if (to_str->len == 0)
    str_sprintf (to_str, "%s", addr->full);
  else
    str_sprintf (to_str, ", %s", addr->full);
}


static void
take_action (void (*fun)(int))
{
  int i;
  int took = 0;

  PREAMBLE;

  /**
   * The reversed order is essential for action_remove.
   */
  for (i = abook->count - 1; i >= 0; i--){
    if (bitarray_is_set (abook_select->selected, i)){
      fun (i);
      took = 1;
    }
  }

  if (! took && abook->count > 0){
    fun (abook_select->bar_pos);
  }
}

/****************************************************************************
 *    ADDRESSBOOK WINDOW INTERFACE FUNCTIONS
 ****************************************************************************/


void
abook_open (void)
{
  rstring_t *items;

  if (! abook_sorted)
    abook_sort ();
  
  items = make_items ();

  abook_select = select_open (items, ABOOK_WIDTH, 20, 1, 0, _("addressbook"));
  if (abook_select == NULL)
    return;

  select_show (abook_select);
  cmd_state_push (CMD_ABOOK);
}



void
abook_close (void)
{
  PREAMBLE;

  rstring_delete (abook_select->items);
  select_close (abook_select);
  abook_select = NULL;
  cmd_state_pop ();
}


void
abook_refresh (void)
{
  PREAMBLE;

  select_show (abook_select);
}



void
abook_next (void)
{
  PREAMBLE;
  
  select_next (abook_select);
}



void
abook_prev (void)
{
  PREAMBLE;
  
  select_prev (abook_select);
}



void
abook_next_page (void)
{
  PREAMBLE;
  
  select_next_page (abook_select);
}



void
abook_prev_page (void)
{
  PREAMBLE;
  
  select_prev_page (abook_select);
}



void
abook_first (void)
{
  PREAMBLE;
  
  select_first (abook_select);
}



void
abook_last (void)
{
  PREAMBLE;
  
  select_last (abook_select);
}


void
abook_hit (void)
{
  PREAMBLE;
  
  select_change (abook_select);
}



void
abook_hit_all (void)
{
  PREAMBLE;
  
  select_change_all (abook_select);
}


void
abook_set (void)
{
  PREAMBLE;
  
  select_set (abook_select);
}


void
abook_unset (void)
{
  PREAMBLE;
  
  select_unset (abook_select);
}


void
abook_set_all (void)
{
  PREAMBLE;
  
  select_set_all (abook_select);
}


void
abook_unset_all (void)
{
  PREAMBLE;
  
  select_unset_all (abook_select);
}



void
abook_remove (void)
{
  rstring_t *items;

  PREAMBLE;

  take_action (action_remove);
  abook_sort ();
  items = make_items ();
  select_install_new_items (abook_select, items);
}


void
abook_change_official (void)
{
  rstring_t *items;
  
  PREAMBLE;

  take_action (action_change_official);
  items = make_items ();
  select_install_new_items (abook_select, items);
}



void
abook_change_sex (void)
{
  rstring_t *items;
  
  PREAMBLE;

  take_action (action_change_sex);
  items = make_items ();
  select_install_new_items (abook_select, items);
}


void
abook_compose (void)
{
  PREAMBLE;

  to_str = str_create ();
  take_action (action_add_to_str);
  abook_close ();
  sender_open_new_to (to_str->str);
  str_destroy (to_str);
  to_str = NULL;
}


void
abook_insert (void)
{
  char      *name;
  char      *email;
  char      *fmt;
  char      *item;
  address_t    *addr;

  PREAMBLE;
  
  name = read_argument ("name: ", NULL, COMPLETE_NONE);
  if (name == NULL)
    return;
  else if (*name == '\0')
    name = NULL;
  else
    name = memblock_strdup (&addr_block, name);

  email = read_argument ("email: ", NULL, COMPLETE_NONE);
  if (email == NULL || (*email == '\0' && name == NULL))
    return;
  else if (*email == '\0')
    email = NULL;
  else
    email = memblock_strdup (&addr_block, email);

  addr                   = address_empty ();
  addr->name             = name;
  addr->email            = email;
  addr->flags.bits.abook = YES;

  address_complete_mem (addr, &addr_block);
  
  raddress_add (abook, addr);

  abook_changed = 1;
  abook_sorted  = 0;

  fmt = ask_for_default ("abook_format", NULL);
  if (fmt == NULL)
    fmt = DEFAULT_FORMAT;
  
  item = eprintf_addr (fmt, addr);

  select_add_item (abook_select, item);
}



void
abook_change_name (void)
{
  address_t  *addr;
  char    *fmt;
  char    *name;
  char   **item;

  PREAMBLE;
  NONEMPTY;
  
  addr = abook->array[abook_select->bar_pos];
  name = read_argument ("name: ", addr->name, COMPLETE_NONE);
  if (name == NULL)
    return;
  else if (*name == '\0')
    name = NULL;
  else
    name = memblock_strdup (&addr_block, name);

  addr->name = name;
  addr->full = NULL;

  address_complete_mem (addr, &addr_block);
  
  fmt = ask_for_default ("abook_format", NULL);
  if (fmt == NULL)
    fmt = DEFAULT_FORMAT;
  
  item  = abook_select->items->array + abook_select->bar_pos;
  xfree (*item);
  *item = eprintf_addr (fmt, addr);

  abook_changed = 1;

  select_redraw (abook_select);
}



void
abook_change_email (void)
{
  address_t     *addr;
  char       *fmt;
  char       *email;
  char      **item;

  PREAMBLE;
  NONEMPTY;

  addr  = abook->array[abook_select->bar_pos];
  email = read_argument ("email: ", addr->email, COMPLETE_NONE);
  
  if (email == NULL)
    return;
  else if (*email == '\0')
    email = NULL;
  else
    email = memblock_strdup (&addr_block, email);
  
  addr->email = email;
  addr->full  = NULL;

  address_complete_mem (addr, &addr_block);

  fmt = ask_for_default ("abook_format", NULL);
  if (fmt == NULL)
    fmt = DEFAULT_FORMAT;
  
  item  = abook_select->items->array + abook_select->bar_pos;
  xfree (*item);
  *item = eprintf_addr (fmt, addr);

  abook_changed = 1;

  select_redraw (abook_select);
}


/****************************************************************************
 *    ADDRESSBOOK ADDING WINDOW FUNCTIONS
 ****************************************************************************/


void
abook_add_open (void)
{
  rstring_t *items;
  mail_t    *mail = folder_mail_selected ();
  
  if (mail == NULL)
    return;
  
  add_raddress = raddress_create ();

  if (mail->from)
    raddress_add (add_raddress, mail->from);
  if (mail->reply_to && mail->reply_to != mail->from)
    raddress_add (add_raddress, mail->reply_to);
  if (mail->to)
    raddress_add_array (add_raddress, mail->to);

  items = raddress_to_strings (add_raddress);

  add_select = select_open (items, COLS - 20, 20, 1, 0, _("select address"));

  if (add_select == NULL)
    return;
  
  select_show (add_select);

  cmd_state_push (CMD_ABOOK_ADD);
}



void
abook_add_close (void)
{
  PREAMBLE_ADD;
  
  if (add_raddress){
    raddress_destroy (add_raddress);
    add_raddress = NULL;
  }

  rstring_delete (add_select->items);
  select_close (add_select);

  add_select = NULL;
  
  cmd_state_pop ();
}



void
abook_add_refresh (void)
{
  PREAMBLE_ADD;

  select_show (add_select);
}



void
abook_add_next (void)
{
  PREAMBLE_ADD;
  
  select_next (add_select);
}



void
abook_add_prev (void)
{
  PREAMBLE_ADD;
  
  select_prev (add_select);
}



void
abook_add_next_page (void)
{
  PREAMBLE_ADD;
  
  select_next_page (add_select);
}



void
abook_add_prev_page (void)
{
  PREAMBLE_ADD;
  
  select_prev_page (add_select);
}



void
abook_add_first (void)
{
  PREAMBLE_ADD;
  
  select_first (add_select);
}



void
abook_add_last (void)
{
  PREAMBLE_ADD;
  
  select_last (add_select);
}



void
abook_add_hit (void)
{
  address_t *addr;

  PREAMBLE_ADD;
  
  if (add_raddress->count == 0)
    return;
  
  addr = add_raddress->array[add_select->bar_pos];
  if (addr->flags.bits.abook)
    return;

  addr = address_copy (addr, &addr_block);
  addr->flags.bits.abook = YES;
  raddress_add (abook, addr);
  abook_changed = 1;
  abook_sorted  = 0;
  select_remove_selected (add_select);
}



/****************************************************************************
 *    CONFIG FUNCTIONS
 ****************************************************************************/


void
abook_new_prepare (void)
{
  memblock_set_mark (addr_block);
  if (new_item.addr == NULL)
    new_item.addr = address_empty_mem (&addr_block);
}



void
abook_new_drop (void)
{
  memblock_free_marked (&addr_block);
  new_item.addr = NULL;
}


/**
 * this function _requires_ that the addressbook is read before any other
 * address is added to address table
 */
void
abook_new_store (void)
{
  address_t *addr;
  
  if (new_item.addr == NULL)
    return;

  new_item.addr->flags.bits.abook = YES;
  if (new_item.addr->name == NULL && new_item.addr->email == NULL){
    abook_new_drop ();
    return;
  }
  
  addr = address_complete_mem (new_item.addr, &addr_block);
  if (addr != new_item.addr){
    abook_new_drop ();
    return;
  }
  raddress_add (abook, addr);
  new_item.addr = NULL;
}



int
abook_new_field (const char *field)
{
  new_item.s_field = NULL;
  new_item.shift   = 0;
  new_item.mask    = 0;
  
  if (strcmp (field, "name") == 0)
    new_item.s_field = &new_item.addr->name;
  else if (strcmp (field, "email") == 0)
    new_item.s_field = &new_item.addr->email;
  else if (strcmp (field, "sex") == 0){
    new_item.shift = A_SHIFT_SEX;
    new_item.mask  = A_MASK_SEX;
  }
  else if (strcmp (field, "initials") == 0)
    new_item.s_field = &new_item.addr->initials;
  else if (strcmp (field, "official") == 0){
    new_item.shift = A_SHIFT_OFFICIAL;
    new_item.mask  = A_MASK_OFFICIAL;
  }
  else if (strcmp (field, "foreign") == 0){
    new_item.shift = A_SHIFT_FOREIGN;
    new_item.mask  = A_MASK_FOREIGN;
  }

  if (new_item.s_field == NULL && new_item.shift == 0
      && new_item.mask == 0)
    return 1;
  return 0;
}



void
abook_new_value (const char *value)
{
  if (addr_block == NULL)
    return;
  
  if (new_item.s_field){
    *new_item.s_field = memblock_strdup (&addr_block, value);
  }
  else if (new_item.mask){
    new_item.addr->flags.value |= (int_value (value) << new_item.shift)
      & new_item.mask;
  }
}



/****************************************************************************
 *    INTERFACE CLASS BODIES
 ****************************************************************************/
/****************************************************************************
 *
 *    END MODULE abook.c
 *
 ****************************************************************************/
