/* 
   elmo - ELectronic Mail Operator

   Copyright (C) 2002, 2003 rzyjontko

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  

   ----------------------------------------------------------------------

   for operating on email addresses
   
*/
/****************************************************************************
 *    IMPLEMENTATION HEADERS
 ****************************************************************************/

#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <regex.h>
#include <ctype.h>

#include "address.h"
#include "ask.h"
#include "xmalloc.h"
#include "memblock.h"
#include "error.h"
#include "misc.h"
#include "hash.h"

/****************************************************************************
 *    IMPLEMENTATION PRIVATE DEFINITIONS / ENUMERATIONS / SIMPLE TYPEDEFS
 ****************************************************************************/

#define INITIAL_BLOCK_SIZE 2048
#define INITIAL_HASH_SIZE  10

/****************************************************************************
 *    REGULAR EXPRESSIONS
 ****************************************************************************/

#define WS "([[:space:]]?)"

#define ATEXT "([[:alpha:][:digit:]!#\\$%&\\\\\\'\\*\\+\\/\\=\\?\\^_\\`" \
              "\\{\\}\\~\\|\\-])"
#define ATOM "(" WS ATEXT "+" WS ")"
#define DOT_ATOM_TEXT "(" ATEXT "+(\\." ATEXT "+)*)"
#define DOT_ATOM "(" DOT_ATOM_TEXT ")"

#define QUOTED_STRING "\"(([^\"])|(\\\\\"))*\""
#define WORD  "(" ATOM "|" QUOTED_STRING ")"
#define PHRASE "(" WORD "+)"


#define ATOM_RE "^" ATOM "$"
#define EMAIL_RE "(" DOT_ATOM "@" DOT_ATOM ")"
#define COMMENT_RE "[(][^()]*[)]"

#define FEMALE_RE "^[^[:space:]]*[aA](([^[:alpha:]])|($))"
#define LAST_RE "[[:space:]][^[:space:]]+$"

/****************************************************************************
 *    IMPLEMENTATION PRIVATE CLASS PROTOTYPES / EXTERNAL CLASS REFERENCES
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE STRUCTURES / UTILITY CLASSES
 ****************************************************************************/

struct table_set {
  int         used;
  memblock_t *data_block;
  htable_t   *address_table;
};

/****************************************************************************
 *    IMPLEMENTATION REQUIRED EXTERNAL REFERENCES (AVOID)
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE DATA
 ****************************************************************************/

static struct table_set descriptors[ADDRESS_DESCRIPTORS_COUNT];

static struct table_set *current = NULL;

static void (*iterator_fun)(address_t *) = NULL;

/****************************************************************************
 *    INTERFACE DATA
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE FUNCTION PROTOTYPES
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE FUNCTIONS
 ****************************************************************************/


/**
 * copy given string to memory block, and remove leading and trailing
 * white spaces
 */
static char *
clean_copy (memblock_t **block, const char *str)
{
  char *last_space = NULL;
  char *end;
  char *result;
  
  while (isspace (*str))
    str++;

  result = memblock_strdup (block, str);
  
  for (end = result; *end; end++){
    if (isspace (*end) && last_space == NULL)
      last_space = end;
    else
      last_space = NULL;
  }

  if (last_space)
    *last_space = '\0';

  return result;
}


/**
 * ALGORITHM:
 *   1. find email (according to RFC2822 syntax) in the addr->full string
 *   2. if it is surrounded with angle brackets (<>), then set kind to
 *      email_in_angles
 *   3. if the type of the addr is still unknown try to find a comment
 *      that can be used as a display name (this is an obsolete form
 *      described in RFC822); when found, set kind to name_as_comment
 *   4. if the email was found (and nothing else), set kind to pure_email
 *   5. set kind to strange
 */
static void
determine_kind (address_t *addr, memblock_t **data_block)
{
  int        len;
  int        off;
  regmatch_t matches[1];

  if (addr->flags.bits.kind != KIND_UNKNOWN)
    return;
  
  if (misc_regex (EMAIL_RE, addr->full, matches) == 0){
    addr->email      = NULL;
    addr->flags.bits.kind = KIND_NO_EMAIL;
  }
  else {
    len         = matches[0].rm_eo - matches[0].rm_so;
    off         = matches[0].rm_so;
    addr->email = memblock_malloc (data_block, len + 1);
    memcpy (addr->email, addr->full + off, len);
    addr->email[len] = '\0';

    if (off > 0 && addr->full[off - 1] == '<'
        && addr->full[off + len] == '>')
      addr->flags.bits.kind = KIND_EMAIL_IN_ANGLES;
    else if (off > 0)
      addr->flags.bits.kind = KIND_STRANGE;

    if (off == 0 && addr->full[len] == '\0')
      addr->flags.bits.kind = KIND_PURE_EMAIL;
  }

  if (addr->flags.bits.kind != KIND_UNKNOWN)
    return;

  if (misc_regex (COMMENT_RE, addr->full, matches)){
    addr->flags.bits.kind = KIND_NAME_AS_COMMENT;
    addr->first           = addr->full + matches[0].rm_so;
  }
  else
    addr->flags.bits.kind = KIND_STRANGE;
}



static void
strip_escapes (char *str)
{
  int len;
  
  if (str == NULL)
    return;

  len = strlen (str);
  
  while (*str){
    if (*str == '\\'){
      memmove (str, str + 1, len);
      len--;
    }
    str++;
    len--;
  }
}


static void
strip_name (address_t *addr)
{
  char *seek;
  char *last;
  
  if (addr->name == NULL)
    return;

  while (isspace (*addr->name) || *addr->name == '(')
    addr->name++;

  last = seek = addr->name;
  while (*seek){
    if (! isspace (*seek) && *seek != ')')
      last = seek;
    seek++;
  }
  seek  = last + 1;
  *seek = '\0';
}



static int
strip_quotes (char *str)
{
  char *src  = str;
  char *dest = str;
  int   off  = 0;
  int   len  = 0;
  int   ret  = 1;
  int   slen;
  regmatch_t matches[1];

  if (str == NULL)
    return 1;

  while (misc_regex (QUOTED_STRING, src, matches)){
    off  = matches[0].rm_so;
    len  = matches[0].rm_eo - matches[0].rm_so;

    memmove (dest, src, off);
    memmove (dest + off, src + off + 1, len - 2);
    dest += off + len - 2;
    src   = src + matches[0].rm_eo;
    ret   = 0;
  }

  if (len){
    slen = strlen (src);
    if (slen > 0)
      memmove (dest, src + 1, slen - 1);
    dest[slen] = '\0';
  }
  return ret;
}



static void
determine_name (address_t *addr, memblock_t **data_block)
{
  int         len;
  int         off;
  char       *re = NULL;
  regmatch_t  matches[1];
  
  if (addr->name != NULL)
    return;
  
  switch (addr->flags.bits.kind){

    case KIND_UNKNOWN:
    case KIND_PURE_EMAIL:
      addr->name = NULL;
      return;

    case KIND_NO_EMAIL:
      addr->name = addr->full;
      return;

    case KIND_EMAIL_IN_ANGLES:
    case KIND_STRANGE:
      re = PHRASE;
      break;

    case KIND_NAME_AS_COMMENT:
      re = COMMENT_RE;
      break;
  }

  if (misc_regex (re, addr->full, matches)){
    off        = matches[0].rm_so;
    len        = matches[0].rm_eo - matches[0].rm_so;
    addr->name = memblock_malloc (data_block, len + 1);
    memcpy (addr->name, addr->full + off, len);
    addr->name[len] = '\0';
  }
  else
    addr->name = NULL;

  strip_name (addr);
  if (strip_quotes (addr->name))
    addr->flags.bits.atomic_name = YES;
  strip_escapes (addr->name);
}



static void
determine_sex (address_t *addr)
{
  char       *str = NULL;
  regmatch_t  matches[1];
  
  if (addr->flags.bits.sex != SEX_UNKNOWN)
    return;
  
  if (addr->first)
    str = addr->first;
  else if (addr->name)
    str = addr->name;
  else if (addr->full)
    str = addr->full;
  else
    str = addr->email;

  if (str == NULL)
    return;
  
  if (misc_regex (FEMALE_RE, str, matches)){
    addr->flags.bits.sex = SEX_FEMALE;
  }
  else {
    addr->flags.bits.sex = SEX_MALE;
  }
}



static void
find_last (address_t *addr)
{
  regmatch_t matches[1];

  if (addr->name == NULL)
    return;
  
  if (misc_regex (LAST_RE, addr->name, matches)){
    addr->last = addr->name + matches[0].rm_so + 1;
  }
}



static void
make_initials (address_t *addr, memblock_t **data_block)
{
  int   first = 1;
  char *seek;
  char *initials;
  char *iseek;

  if (addr->name == NULL)
    return;

  if (*addr->name == '\0')
    return;
  
  iseek = initials = xmalloc (strlen (addr->name));

  for (seek = addr->name; *seek; seek++){
    if (isalpha (*seek) && first){
      *iseek = *seek;
      first  = 0;
      iseek++;
    }
    else if (isspace (*seek)){
      first  = 1;
    }
  }
  *iseek = '\0';
  addr->initials = memblock_strdup (data_block, initials);
  xfree (initials);
}


static int
is_atom (const char *str)
{
  regmatch_t matches[1];
  
  return misc_regex (ATOM_RE, str, matches);
}


static void
make_full (address_t *addr, memblock_t **data_block)
{
  int   ret;
  int   len;
  int   n_len;
  int   e_len;
  int   f_len;
  char *f_ptr;
  char *seek;
  
  if (addr->name && addr->email){
    ret = is_atom (addr->name);
    if (ret){
      addr->flags.bits.atomic_name = YES;
      addr->full = memblock_sprintf (data_block, "%s <%s>", addr->name,
                                     addr->email);
      return;
    }

    addr->flags.bits.atomic_name = NO;
    
    n_len = strlen (addr->name);
    e_len = strlen (addr->email);
    /*      "    name       "   <   email   >  \0 */
    f_len = 1 + 2 * n_len + 1 + 1 + e_len + 1 + 1;
    
    f_ptr  = addr->full = memblock_malloc (data_block, f_len);
    *f_ptr = '"';
    f_ptr++;
    for (seek = addr->name; *seek; seek++){
      if (*seek == '"'){
        *f_ptr = '\\';
        f_ptr++;
        addr->flags.bits.quotes_in_name = YES;
      }
      else if (*seek == '\\'){
        *f_ptr = '\\';
        f_ptr++;
      }
      *f_ptr = *seek;
      f_ptr++;
    }

    len  = f_ptr - addr->full;
    len += sprintf (f_ptr, "\" <%s>", addr->email);
    len += 1;

    memblock_shrink_last (*data_block, len);
  }
  else if (addr->name){
    addr->full = addr->name;
  }
  else {
    addr->full = addr->email;
  }
}



static void
complete_info (address_t *addr, memblock_t **data_block)
{
  if (addr->full == NULL)
    make_full (addr, data_block);
  if (addr->flags.bits.kind == KIND_UNKNOWN)
    determine_kind (addr, data_block);
  if (addr->name == NULL)
    determine_name (addr, data_block);
  if (addr->flags.bits.sex == SEX_UNKNOWN)
    determine_sex (addr);
  if (addr->last == NULL)
    find_last (addr);
  if (addr->initials == NULL)
    make_initials (addr, data_block);
}


static address_t *
insert (address_t *addr)
{
  entry_t *entry;

  entry = htable_insert (current->address_table, addr->full, NULL);
  if (entry->content != NULL)
    return entry->content;

  entry->content = addr;
  return NULL;
}


static void
iterator (entry_t *entry)
{
  address_t *addr = entry->content;

  iterator_fun (addr);
}


static void
replace_in_hash (address_t *addr)
{
  entry_t *entry;

  entry = htable_insert (current->address_table, addr->full, NULL);
  if (entry->key != addr->full){
    entry->key = addr->full;
  }
  entry->content = addr;
}


/****************************************************************************
 *    INTERFACE FUNCTIONS
 ****************************************************************************/


void
address_init (void)
{
  memset (descriptors, '\0', sizeof (descriptors));
}



void
address_free_resources (void)
{
  int i;
  struct table_set *tset;

  for (i = 0; i < ADDRESS_DESCRIPTORS_COUNT; i++){
    tset = descriptors + i;
    if (tset->used){
      if (tset->data_block)
        memblock_destroy (tset->data_block);
      if (tset->address_table)
        htable_destroy (tset->address_table, NULL);
    }
  }
  current = NULL;
}


/****************************************************************************
 *    OBJECT OPERATIONS
 ****************************************************************************/

address_t *
address_empty_mem (memblock_t **data_block)
{
  address_t *addr;

  if (! current || ! current->used)
    return NULL;

  addr = memblock_malloc (data_block, sizeof (address_t));

  addr->full        = NULL;
  addr->email       = NULL;
  addr->name        = NULL;
  addr->first       = NULL;
  addr->last        = NULL;
  addr->initials    = NULL;
  addr->flags.value = 0;
  
  return addr;
}


address_t *
address_empty (void)
{
  return address_empty_mem (&current->data_block);
}


address_t *
address_from_string_mem (const char *str, memblock_t **data_block)
{
  address_t *addr;
  address_t *inserted;

  if (! current || ! current->used)
    return NULL;
  
  memblock_set_mark (*data_block);
  addr       = address_empty_mem (data_block);
  addr->full = clean_copy (data_block, str);
  inserted   = insert (addr);

  if (inserted){
    memblock_free_marked (data_block);
    return inserted;
  }

  complete_info (addr, data_block);
  return addr;
}


address_t *
address_from_string (const char *str)
{
  return address_from_string_mem (str, &current->data_block);
}


address_t *
address_complete_mem (address_t *addr, memblock_t **data_block)
{
  address_t *inserted;
  
  complete_info (addr, data_block);

  inserted = insert (addr);

  if (inserted){
    return inserted;
  }

  return addr;
}


address_t *
address_complete (address_t *addr)
{
  return address_complete_mem (addr, &current->data_block);
}


address_t *
address_copy (address_t *addr, memblock_t **data_block)
{
  address_t *result = memblock_malloc (data_block, sizeof (address_t));
  
  if (addr->full)
    result->full = memblock_strdup (data_block, addr->full);
  if (addr->email)
    result->email = memblock_strdup (data_block, addr->email);
  if (addr->name)
    result->name = memblock_strdup (data_block, addr->name);
  if (addr->first)
    result->first = memblock_strdup (data_block, addr->first);
  if (addr->last)
    result->last = memblock_strdup (data_block, addr->last);
  if (addr->initials)
    result->initials = memblock_strdup (data_block, addr->initials);
  result->flags.value = addr->flags.value;

  replace_in_hash (result);
  return result;
}


/****************************************************************************
 *    SERVICE FUNCTIONS
 ****************************************************************************/


char *
address_name (address_t *addr)
{
  if (addr == NULL)
    return NULL;

  if (addr->name)
    return addr->name;

  return addr->full;
}


char *
address_wrote_format (address_t *addr)
{
  if (addr == NULL)
    return NULL;
  
  switch (addr->flags.bits.sex){

    case SEX_UNKNOWN:
      return NULL;

    case SEX_MALE:
      if (addr->flags.bits.official)
        return ask_for_default ("o_male_wrote", NULL);
      else
        return ask_for_default ("male_wrote", NULL);

    case SEX_FEMALE:
      if (addr->flags.bits.official)
        return ask_for_default ("o_female_wrote", NULL);
      else
        return ask_for_default ("female_wrote", NULL);

    default:
      return NULL;
  }
}


int
address_cmp (const address_t *a, const address_t *b)
{
  char *as = NULL;
  char *bs = NULL;

  if (a->last)
    as = a->last;
  else if (a->name)
    as = a->name;
  else
    as = a->full;

  if (b->last)
    bs = b->last;
  else if (b->name)
    bs = b->name;
  else
    bs = b->full;

  if (as == NULL || bs == NULL)
    return 0;

  return strcmp (as, bs);
}



int
address_open (void)
{
  int i;
  struct table_set *tset;

  for (i = 0; i < ADDRESS_DESCRIPTORS_COUNT; i++){
    tset = descriptors + i;
    if (! tset->used)
      break;
  }
  if (i == ADDRESS_DESCRIPTORS_COUNT)
    return -1;

  tset->data_block    = NULL;
  tset->address_table = NULL;
  tset->used          = 1;

  return i;
}



void
address_close (int ad)
{
  struct table_set *tset;

  if (ad < 0 || ad >= ADDRESS_DESCRIPTORS_COUNT)
    return;
  
  tset = descriptors + ad;

  if (! tset->used)
    return;

  if (tset->data_block){
    memblock_destroy (tset->data_block);
    tset->data_block = NULL;
  }
  if (tset->address_table){
    htable_destroy (tset->address_table, NULL);
    tset->address_table = NULL;
  }
  tset->used = 0;
}


void
address_point (int ad)
{
  if (ad < 0 || ad >= ADDRESS_DESCRIPTORS_COUNT)
    return;
  
  current = descriptors + ad;

  if (current->data_block == NULL)
    current->data_block    = memblock_create (INITIAL_BLOCK_SIZE);
  if (current->address_table == NULL)
    current->address_table = htable_create (INITIAL_HASH_SIZE);
}



void
address_for_all (int ad, void (*fun)(address_t *))
{
  struct table_set *tset;
  
  if (ad < 0 || ad >= ADDRESS_DESCRIPTORS_COUNT)
    return;
  
  tset = descriptors + ad;
  
  if (tset->address_table){
    iterator_fun = fun;
    htable_iterator (tset->address_table, iterator);
    iterator_fun = NULL;
  }
}



int
address_count (int ad)
{
  struct table_set *tset;
  
  if (ad < 0 || ad >= ADDRESS_DESCRIPTORS_COUNT)
    return -1;

  tset = descriptors + ad;

  if (tset->address_table)
    return tset->address_table->count;

  return 0;
}


/****************************************************************************
 *    INTERFACE CLASS BODIES
 ****************************************************************************/
/****************************************************************************
 *
 *    END MODULE address.c
 *
 ****************************************************************************/
