/* 
   elmo - ELectronic Mail Operator

   Copyright (C) 2003 rzyjontko

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  

   ----------------------------------------------------------------------

   Different sring manipulation routines.
   
*/
/****************************************************************************
 *    IMPLEMENTATION HEADERS
 ****************************************************************************/

#include <stdarg.h>
#include <string.h>
#include <ctype.h>

#include "mail.h"
#include "xmalloc.h"
#include "wrapbox.h"
#include "misc.h"
#include "keymap.h"

/****************************************************************************
 *    IMPLEMENTATION PRIVATE DEFINITIONS / ENUMERATIONS / SIMPLE TYPEDEFS
 ****************************************************************************/

#define FIRST_ALLOC 20

enum justify { JUSTIFY_LEFT = 0, JUSTIFY_RIGHT = 1};

/****************************************************************************
 *    IMPLEMENTATION PRIVATE CLASS PROTOTYPES / EXTERNAL CLASS REFERENCES
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE STRUCTURES / UTILITY CLASSES
 ****************************************************************************/

typedef struct {
  char *str;
  char *end;
  int   size;
} str_t;

/**
 * This structure describes formatting sequence (i.e. %-020n)
 * everything between % and the first letter is used to format
 * a string.
 * In the example above -020 means that we want the desired string
 * be exactly 20 characters long, filled with spaces (leading 0),
 * and justified to right (dash).
 *
 * To be more precise:
 *    len       = 20
 *    character = 'n'
 *    fill      = YES
 *    justify   = JUSTIFY_RIGHT
 */
struct fmt_seq {
  int  fs_len;
  
  int  len;
  char character;
  struct {
    enum yes_no  fill    : 1;
    enum justify justify : 1;
  } flags;
};

/****************************************************************************
 *    IMPLEMENTATION REQUIRED EXTERNAL REFERENCES (AVOID)
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE DATA
 ****************************************************************************/
/****************************************************************************
 *    INTERFACE DATA
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE FUNCTION PROTOTYPES
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE FUNCTIONS
 ****************************************************************************/

static void
eval_fmt_seq (struct fmt_seq *fs, const char *s)
{
  char *rest;
  
  fs->fs_len        = 0;
  fs->len           = 0;
  fs->character     = '\0';
  fs->flags.fill    = NO;
  fs->flags.justify = JUSTIFY_LEFT;

  if (*s != '%')
    return;

  s++;

  if (*s == '-'){
    fs->flags.justify = JUSTIFY_RIGHT;
    s++;
    fs->fs_len++;
  }

  if (*s == '0'){
    fs->flags.fill = YES;
    s++;
    fs->fs_len++;
  }

  if (isdigit (*s)){
    fs->len = strtol (s, &rest, 10);
    if (rest)
      fs->fs_len += rest - s;
    s = rest;
  }

  fs->character = *s;
  fs->fs_len++;
}


static void
initialize_str (str_t *s)
{
  s->size = FIRST_ALLOC;
  s->str  = xmalloc (s->size);
  s->end  = s->str;
  *s->end = '\0';
}


static void
resize_str (str_t *s)
{
  int len = s->end - s->str;
  
  if (len == s->size - 1){
    s->size = (s->size + 1) * 2;
    s->str  = xrealloc (s->str, s->size);
    s->end  = s->str + len;
  }
}


static void
put_char (str_t *s, char c)
{
  resize_str (s);
  *s->end = c;
  s->end++;
  *s->end = '\0';
}


static void
put_nchar (str_t *s, char c, int n)
{
  while (n){
    put_char (s, c);
    n--;
  }
}



static void
put_string_len (str_t *s, const char *t, int len)
{
  int oldlen = s->end - s->str;
  
  if (!t)
    return;

  if (len + oldlen >= s->size - 1){
    s->size = (s->size + len) * 2;
    s->str  = xrealloc (s->str, s->size);
  }

  memcpy (s->str + oldlen, t, len);
  s->str[oldlen + len] = '\0';
  s->end = s->str + oldlen + len;
}



static void
put_string (str_t *s, const char *t)
{
  int len;

  len = strlen (t);
  put_string_len (s, t, len);
}



static void
put_string_fs (str_t *s, const char *t, struct fmt_seq *fs)
{
  int spaces;
  int t_len;
  int w_len;
  
  if (! t)
    t = "";

  t_len = strlen (t);

  if (t_len <= fs->len){
    w_len  = t_len;
    spaces = fs->len - t_len;
  }
  else if (fs->len != 0){
    w_len  = fs->len;
    spaces = 0;
  }
  else {
    w_len  = t_len;
    spaces = 0;
  }

  if (fs->flags.justify == JUSTIFY_LEFT){
    put_string_len (s, t, w_len);
    if (fs->flags.fill)
      put_nchar (s, ' ', spaces);
  }
  else {
    if (fs->flags.fill)
      put_nchar (s, ' ', spaces);
    put_string_len (s, t, w_len);
  }
}


static void
put_int (str_t *s, int d)
{
  static char buf[15];

  sprintf (buf, "%d", d);
  put_string (s, buf);
}


/**
 * Suppose a mail:
 *  From: Johnny Masta Killa <johnny@masta.com>
 *  To: Bob The Ripper <bob@ripper.com>
 *  Subject: Bow before me!
 *  Date: Wed, 7 May 2003 15:36:46 +0200
 *
 */
static int
put_mail_special (str_t *s, const char *fmt, mail_t *mail)
{
  int   size;
  char *date;
  struct fmt_seq fs;

  eval_fmt_seq (&fs, fmt);
  
  resize_str (s);
  switch (fs.character){

    case '%':
      put_char (s, '%');
      break;

      /* Johnny Masta Killa */
    case 'f':
      put_string_fs (s, address_name (mail->from), &fs);
      break;

      /* Johnny Masta Killa <johnny@masta.com> */
    case 'F':
      if (mail->from)
        put_string_fs (s, mail->from->full, &fs);
      else
        put_string_fs (s, NULL, &fs);
      break;

      /* Bob The Ripper */
    case 't':
      if (mail->to)
        put_string_fs (s, address_name (mail->to->array[0]), &fs);
      else
        put_string_fs (s, NULL, &fs);
      break;
      
      /* Bob The Ripper <bob@ripper.com> */
    case 'T':
      if (mail->to && mail->to->count)
        put_string_fs (s, mail->to->array[0]->full, &fs);
      else
        put_string_fs (s, NULL, &fs);
      break;

      /* Wed, 7 May 2003 15:36:46 +0200 */
    case 'd':
      put_string_fs (s, mail->date_str, &fs);
      break;

      /* 07 May */
    case 'D':
      date = date_string ("%b %d", mail->date);
      put_string_fs (s, date, &fs);
      xfree (date);
      break;

      /* Bow before me! */
    case 's':
      put_string_fs (s, mail->subject, &fs);
      break;

      /* JMK */
    case 'i':
      if (mail->from)
        put_string_fs (s, mail->from->initials, &fs);
      else
        put_string_fs (s, NULL, &fs);
      break;

      /* (3.5K) */
    case 'S':
      size = wrapbox_mail_size (mail);
      put_string_fs (s, human_size_pure (size), &fs);
      break;

      /* these are flags */
    case '$':
      if (mail->flags & FLAG_FLAGGED)
        put_char (s, 'F');
      else if (mail->flags & FLAG_READ)
        put_char (s, ' ');
      else if (mail->flags & FLAG_OLD)
        put_char (s, 'O');
      else
        put_char (s, 'N');

      if (mail->flags & FLAG_TRASHED)
        put_char (s, 'T');
      else if (mail->flags & FLAG_ANSWERED)
        put_char (s, 'r');
      else if (mail->flags & FLAG_PASSED)
        put_char (s, 'f');
      else
        put_char (s, ' ');
      break;

      /* attachments indicator */
    case '?':
      if (mail->rmime->count > 1)
        put_char (s, '+');
      else
        put_char (s, ' ');
      break;

      /* spam indicator */
    case '#':
      if (mail->flags & FLAG_SPAM)
        put_char (s, '#');
      else if (mail->flags & FLAG_LEGITIMATE
               || mail->flags & FLAG_NOSPAM)
        put_char (s, ' ');
      else
        put_char (s, '?');
      break;
  }

  return fs.fs_len;
}


static int
put_editor_special (str_t *s, const char *fmt, const char *fname, int lineno)
{
  struct fmt_seq fs;

  eval_fmt_seq (&fs, fmt);

  resize_str (s);
  switch (fs.character){

    case '%':
      put_char (s, '%');
      break;

    case 'f':
      put_string_fs (s, fname, &fs);
      break;

    case 'd':
      put_int (s, lineno);
      break;
  }

  return fs.fs_len;
}


static int
put_addr_special (str_t *s, const char *fmt, address_t *addr)
{
  struct fmt_seq fs;

  eval_fmt_seq (&fs, fmt);
  
  resize_str (s);
  switch (fs.character){

    case '%':
      put_char (s, '%');
      break;

    case 'n':
      put_string_fs (s, addr->name, &fs);
      break;

    case 'e':
      put_string_fs (s, addr->email, &fs);
      break;

    case 'o':
      if (addr->flags.bits.official)
        put_char (s, 'o');
      else
        put_char (s, ' ');
      break;

    case 'f':
      if (addr->flags.bits.foreign)
        put_char (s, 'f');
      else
        put_char (s, ' ');
      break;

    case 's':
      if (addr->flags.bits.sex == SEX_MALE)
        put_char (s, 'M');
      else if (addr->flags.bits.sex == SEX_FEMALE)
        put_char (s, 'F');
      else
        put_char (s, '?');
      break;
  }
  return fs.fs_len;
}


static int
put_key_special (str_t *s, const char *fmt, struct key *key)
{
  struct fmt_seq fs;

  eval_fmt_seq (&fs, fmt);
  
  resize_str (s);
  switch (fs.character){

    case '%':
      put_char (s, '%');
      break;

    case 'k':
      put_string_fs (s, key2string (key->key), &fs);
      break;

    case 'f':
      put_string_fs (s, key->exec->name, &fs);
      break;

    case 'd':
      put_string_fs (s, key->exec->desc, &fs);
      break;
  }

  return fs.fs_len;
}


static char *
substitute_mail (const char *fmt, mail_t *mail, char *buf)
{
  str_t       result;
  int         fs_len;
  const char *seek;

  if (buf){
    result.str  = buf;
    result.end  = buf;
    result.size = 1000000000;
    *result.end = '\0';
  }
  else {
    initialize_str (&result);
  }

  for (seek = fmt; *seek; seek++){
    if (*seek == '%'){
      fs_len = put_mail_special (&result, seek, mail);
      seek  += fs_len;
    }
    else
      put_char (&result, *seek);
  }

  return result.str;
}



static char *
substitute_editor (const char *fmt, const char *fname, int lineno)
{
  str_t       result;
  int         fs_len;
  const char *seek;

  initialize_str (&result);
  for (seek = fmt; *seek; seek++){
    if (*seek == '%'){
      fs_len = put_editor_special (&result, seek, fname, lineno);
      seek  += fs_len;
    }
    else
      put_char (&result, *seek);
  }

  return result.str;
}


static char *
substitute_addr (const char *fmt, address_t *addr)
{
  str_t       result;
  int         fs_len;
  const char *seek;

  initialize_str (&result);
  for (seek = fmt; *seek; seek++){
    if (*seek == '%'){
      fs_len = put_addr_special (&result, seek, addr);
      seek  += fs_len;
    }
    else
      put_char (&result, *seek);
  }

  return result.str;
}


static char *
substitute_key (const char *fmt, struct key *key)
{
  str_t       result;
  int         fs_len;
  const char *seek;

  initialize_str (&result);
  for (seek = fmt; *seek; seek++){
    if (*seek == '%'){
      fs_len = put_key_special (&result, seek, key);
      seek  += fs_len;
    }
    else
      put_char (&result, *seek);
  }

  return result.str;
}


/****************************************************************************
 *    INTERFACE FUNCTIONS
 ****************************************************************************/

char *
eprintf_mail (const char *fmt, mail_t *mail, char *buf)
{
  char    *result;

  result = substitute_mail (fmt, mail, buf);
  return result;
}



char *
eprintf_editor (const char *fmt, const char *fname, int lineno)
{
  char *result;

  result = substitute_editor (fmt, fname, lineno);
  return result;
}


char *
eprintf_addr (const char *fmt, address_t *addr)
{
  char *result;
  
  result = substitute_addr (fmt, addr);
  return result;
}


char *
eprintf_key (const char *fmt, struct key *key)
{
  char *result;

  result = substitute_key (fmt, key);
  return result;
}

/****************************************************************************
 *    INTERFACE CLASS BODIES
 ****************************************************************************/
/****************************************************************************
 *
 *    END MODULE eprintf.c
 *
 ****************************************************************************/
