/* 
   elmo - ELectronic Mail Operator 

   Copyright (C) 2003 rzyjontko

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  

   ----------------------------------------------------------------------

   error display system

   An error message must be visible.  It has to remain on the screen
   sufficiently long, so that a user is able to read it.  The message
   should then disappear.  Sometimes one error may generate next errors.
   This is why messages are queued for display.
   
*/

#define _GNU_SOURCE 1

/****************************************************************************
 *    IMPLEMENTATION HEADERS
 ****************************************************************************/

#include <pthread.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <unistd.h>
#include <sys/types.h>
#include <time.h>

#include "ecurses.h"
#include "error.h"
#include "xmalloc.h"
#include "str.h"
#include "cmd.h"
#include "elmo.h"
#include "screen.h"

/****************************************************************************
 *    IMPLEMENTATION PRIVATE DEFINITIONS / ENUMERATIONS / SIMPLE TYPEDEFS
 ****************************************************************************/

#define DELAY  1
#define REMAIN 4

/****************************************************************************
 *    IMPLEMENTATION PRIVATE CLASS PROTOTYPES / EXTERNAL CLASS REFERENCES
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE STRUCTURES / UTILITY CLASSES
 ****************************************************************************/

struct list {
  struct list *next;
  char        *msg;
};

struct queue {
  struct list     *list;
  struct list     *last;
  pthread_mutex_t  mutex;
  time_t           time;
  int              visible;
};

/****************************************************************************
 *    IMPLEMENTATION REQUIRED EXTERNAL REFERENCES (AVOID)
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE DATA
 ****************************************************************************/

/**
 * This queue holds messages to be displayed.
 */
static struct queue msg_queue;

/**
 * This mutex is used to prevent paralell strerror executions because
 * strerror is not a reentrant procedure - it writes an error message
 * to a static buffer, and returns it.
 */
static pthread_mutex_t strerror_mutex;

/**
 * This is an error window, but it is not created here, but points to
 * cmd_echo_area.
 */
static WINDOW *error_window = NULL;

/****************************************************************************
 *    INTERFACE DATA
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE FUNCTION PROTOTYPES
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE FUNCTIONS
 ****************************************************************************/

static void
queue_insert (char *msg)
{
  struct list *elem = xmalloc (sizeof (struct list));
  
  elem->next = NULL;
  elem->msg  = msg;
  
  pthread_mutex_lock (&msg_queue.mutex);
  if (msg_queue.last == NULL){
    msg_queue.list = msg_queue.last = elem;
  }
  else {
    msg_queue.last->next = elem;
    msg_queue.last       = elem;
  }
  pthread_mutex_unlock (&msg_queue.mutex);
}


static char *
queue_get (void)
{
  char        *result;
  struct list *tmp;
  
  if (msg_queue.list == NULL)
    return NULL;

  tmp    = msg_queue.list;
  result = tmp->msg;

  msg_queue.list = tmp->next;
  if (msg_queue.list == NULL)
    msg_queue.last = NULL;

  xfree (tmp);
  return result;
}



static char *
prepare_error_msg (int errnum, const char *fmt, va_list ap)
{
  str_t *str = str_create ();

  str_vsprintf (str, fmt, ap);
  
  if (errnum){
    pthread_mutex_lock (&strerror_mutex);
    str_sprintf (str, ": %s", strerror (errnum));
    pthread_mutex_unlock (&strerror_mutex);
  }

  return str_finished (str);
}


static void
display_msg (const char *msg, int color)
{
  wbkgd (error_window, color);
  mvwaddstr (error_window, 0, 0, msg);
  wclrtoeol (error_window);
  screen_touchwin (error_window);
  screen_wnoutrefresh (error_window);
}

/****************************************************************************
 *    INTERFACE FUNCTIONS
 ****************************************************************************/

void
error_init (void)
{
  error_window = cmd_echo_area;

  memset (&msg_queue, '\0', sizeof (msg_queue));
  pthread_mutex_init (&msg_queue.mutex, NULL);
  pthread_mutex_init (&strerror_mutex, NULL);
}




void
error_hook (void)
{
  char   *msg;
  time_t  now = time (NULL);

  pthread_mutex_lock (&msg_queue.mutex);
  if (msg_queue.visible || msg_queue.list){
    if (now - msg_queue.time > DELAY){
      msg = queue_get ();
      if (msg){
        display_msg (msg, COLOR_PAIR (4));
        msg_queue.time    = now;
        msg_queue.visible = 1;
        xfree (msg);
      }
    }

    if (now - msg_queue.time > REMAIN){
      display_msg ("", 0);
      msg_queue.visible = 0;
    }
  }
  pthread_mutex_unlock (&msg_queue.mutex);
}



void
error_critical (int status, int errnum, const char *msg)
{
  elmo_finish (0);
  fprintf (stderr, "%s: %s\n", msg, strerror (errnum));
  exit (status);
}



void
error_ (int errnum, const char *fmt, ...)
{
  va_list  ap;
  char    *msg;

  va_start (ap, fmt);
  msg = prepare_error_msg (errnum, fmt, ap);
  va_end (ap);

  queue_insert (msg);
  error_hook ();
}



void
error_regex (int errcode, regex_t *compiled)
{
  int   length = regerror (errcode, compiled, NULL, 0);
  char *buffer = xmalloc (length + 1);

  regerror (errcode, compiled, buffer, length);

  queue_insert (buffer);
  error_hook ();
}



void
error_vmsg (const char *fmt, va_list ap)
{
  char *msg;

  msg = prepare_error_msg (0, fmt, ap);

  display_msg (msg, COLOR_PAIR (7));
  xfree (msg);
}



void
error_msg (const char *fmt, ...)
{
  va_list ap;

  va_start (ap, fmt);
  error_vmsg (fmt, ap);
  va_end (ap);
}



void
error_free_resources (void)
{
  struct list *list;
  struct list *next;

  for (list = msg_queue.list; list; list = next){
    next = list->next;
    xfree (list->msg);
    xfree (list);
  }
}



void
error_clear_force (void)
{
  display_msg ("", 0);
}


/****************************************************************************
 *    INTERFACE CLASS BODIES
 ****************************************************************************/
/****************************************************************************
 *
 *    END MODULE error.c
 *
 ****************************************************************************/
