/* 
   elmo - ELectronic Mail Operator

   Copyright (C) 2003 rzyjontko

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  

   ----------------------------------------------------------------------

   mail fetching window
   
*/
/****************************************************************************
 *    IMPLEMENTATION HEADERS
 ****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/mman.h>

#include "ecurses.h"
#include "pop.h"
#include "error.h"
#include "select.h"
#include "line.h"
#include "xmalloc.h"
#include "wrapbox.h"
#include "misc.h"
#include "cmd.h"
#include "file.h"
#include "ask.h"
#include "lcs.h"
#include "fetch.h"
#include "gettext.h"
#include "mybox.h"
#include "eprintf.h"

/****************************************************************************
 *    IMPLEMENTATION PRIVATE DEFINITIONS / ENUMERATIONS / SIMPLE TYPEDEFS
 ****************************************************************************/

#define WINDOW_WIDTH COLS - 6

#define PREAMBLE do { if (fetch_select == NULL) return; } while (0)

/****************************************************************************
 *    IMPLEMENTATION PRIVATE CLASS PROTOTYPES / EXTERNAL CLASS REFERENCES
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE STRUCTURES / UTILITY CLASSES
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION REQUIRED EXTERNAL REFERENCES (AVOID)
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE DATA
 ****************************************************************************/

static select_t *fetch_select = NULL;

static char *host = NULL;


static int         prev_stat_size = 0;
static pop_stat_t *prev_stat_list = NULL;

/**
 * ask opened in begin_connection must be closed either after window close
 * or after begin_connection
 */
static ask_t *fetch_ask = NULL;

/**
 * This is the default format of the line displayed in the fetch window.
 */
static char *default_format = "%?%$  %D %016f (%-06S)  %s";

/****************************************************************************
 *    INTERFACE DATA
 ****************************************************************************/

int fetch_got_mail = 0;

/****************************************************************************
 *    IMPLEMENTATION PRIVATE FUNCTION PROTOTYPES
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE FUNCTIONS
 ****************************************************************************/


static rstring_t *
make_items (const char *fmt)
{
  int        i;
  rstring_t *result = rstring_create ();

  for (i = 0; i < pop_stat_size; i++){
    if (pop_stat_list[i].mark == 1)
      continue;
    rstring_add (result, eprintf_mail (fmt, pop_stat_list[i].info, NULL));
  }
  return result;
}



static int
stats_equal (void *a, void *b)
{
  pop_stat_t *aa = (pop_stat_t *) a;
  pop_stat_t *bb = (pop_stat_t *) b;

  return aa->size == bb->size;
}



static void
mark_stat (void *a, void *b)
{
  pop_stat_t *aa = (pop_stat_t *) a;
  pop_stat_t *bb = (pop_stat_t *) b;

  if (bb->mark == 1)
    aa->mark = 1;
}



static int
selected_index (void)
{
  int i;
  int index = 0;

  for (i = 0; i < pop_stat_size; i++){
    if (pop_stat_list[i].mark == 0 || pop_stat_list[i].fetched)
      index++;
    if (index > fetch_select->bar_pos)
      break;
  }
  if (i < pop_stat_size)
    return i;
  return -1;
}


static int
selected_num (void)
{
  int index = selected_index ();

  if (index == -1)
    return -1;
  return pop_stat_list[index].num;
}



static int
begin_connection (void)
{
  int    port = 0;
  char  *s_port;
  char  *user;
  char  *pass;

  fetch_ask = ask_select ("pop_acc");

  if (fetch_ask == NULL)
    return 1;

  host   = ask_get_field (fetch_ask, "server");
  user   = ask_get_field (fetch_ask, "username");
  pass   = ask_get_field (fetch_ask, "password");
  s_port = ask_get_field (fetch_ask, "port");

  if (s_port)
    port = atoi (s_port);

  if (host == NULL){
    error_ (0, _("server field in pop_acc not defined"));
    ask_destroy (fetch_ask);
    fetch_ask = NULL;
    return 1;
  }

  if (user == NULL){
    error_ (0, _("username field in pop_acc not defined"));
    ask_destroy (fetch_ask);
    fetch_ask = NULL;
    return 1;
  }

  if (pass == NULL){
    error_ (0, _("password field in pop_acc not defined"));
    ask_destroy (fetch_ask);
    fetch_ask = NULL;
    return 1;
  }
  
  fetch_got_mail = 0;

  if (pop_begin_connection (host, (unsigned short) port)){
    ask_destroy (fetch_ask);
    fetch_ask = NULL;
    return 1;
  }

  if (pop_authorize (user, pass)){
    ask_destroy (fetch_ask);
    fetch_ask = NULL;
    return 1;
  }
  
  if (pop_list ())
    return 1;

  return 0;
}



/**
 * !!! this function takes the index in pop_stat_list !!!
 */
static int
get_mail (int index)
{
  int     msg_index;
  int     fd;
  char   *fname;
  FILE   *fp;

  if (pop_stat_list[index].fetched)
    return 1;

  msg_index = pop_stat_list[index].num;
  
  fname = wrapbox_fetch_where ();
  fd    = open (fname, O_RDWR | O_CREAT | O_EXCL, 0600);
  if (fd == -1){
    error_ (errno, _("couldn't write to %s"), fname);
    return 1;
  }

  fp = fdopen (fd, "w");
  if (!fp){
    error_ (errno, "%s", fname);
    return 1;
  }

  if (pop_retr (msg_index, fp)){
    error_ (0, _("couldn't fetch mail %d"), msg_index);
    fclose (fp);
    xfree (fname);
    return 1;
  }

  fclose (fp);

  wrapbox_deliver_to (fname, wrapbox_marray->path);
  xfree (fname);

  return 0;
}



static void
no_new_mail (rstring_t *items)
{
  ask_destroy (fetch_ask);
  fetch_ask = NULL;
  fetch_remove_till_limit ();
  pop_finish_connection ();
  rstring_free_strings (items);
  rstring_delete (items);
  host = NULL;
  error_ (0, _("no new mail"));
  cmd_refresh ();
}



static void
open_window (void)
{
  char      *fmt = ask_for_default ("fetch_format", NULL);
  rstring_t *items;

  if (pop_list_ext ())
    return;

  if (fmt == NULL)
    fmt = default_format;

  items = make_items (fmt);
  if (items->count == 0){
    fetch_save_pop_list ();
    no_new_mail (items);
    pop_finish_connection ();
    return;
  }

  fetch_select = select_open (items, WINDOW_WIDTH, 20, 1, 0, _("fetch"));

  if (fetch_select){
    select_show (fetch_select);
    cmd_state_push (CMD_FETCH);
  }
}



/****************************************************************************
 *    INTERFACE FUNCTIONS
 ****************************************************************************/


void
fetch_all (void)
{
  int i;
  
  if (begin_connection ())
    return;

  if (fetch_ask)
    ask_destroy (fetch_ask);
  fetch_ask = NULL;

  fetch_load_pop_list ();
  fetch_merge_lists ();
  
  for (i = 0; i < pop_stat_size; i++){
    if (pop_stat_list[i].mark == 1)
      continue;
    if (get_mail (i) == 0)
      pop_stat_list[i].mark = 1;
  }

  fetch_save_pop_list ();
  fetch_remove_till_limit ();
  fetch_remove_till_count ();

  host = NULL;
  
  pop_finish_connection ();
}




void
fetch_get_rest (void)
{
  int        i;
  char      *fmt;
  rstring_t *items;
  
  PREAMBLE;

  for (i = 0; i < pop_stat_size; i++){
    pop_stat_list[i].mark = 0;
  }

  if (pop_list_ext ())
    return;

  fmt = ask_for_default ("fetch_format", NULL);
  if (fmt == NULL)
    fmt = default_format;

  items = make_items (fmt);
  if (items->count == 0){
    no_new_mail (items);
    fetch_close ();
    return;
  }

  select_install_new_items (fetch_select, items);
}



void
fetch_open (void)
{
  if (begin_connection ())
    return;
  
  fetch_load_pop_list ();
  fetch_merge_lists ();
  
  open_window ();
}



void
fetch_open_all (void)
{
  if (begin_connection ())
    return;
  
  open_window ();
}



void
fetch_close (void)
{
  PREAMBLE;

  fetch_save_pop_list ();
  fetch_remove_till_limit ();
  fetch_remove_till_count ();

  host = NULL;
  
  pop_finish_connection ();
  rstring_free_strings (fetch_select->items);
  rstring_delete (fetch_select->items);
  select_close (fetch_select);
  fetch_select = NULL;

  cmd_state_pop ();

  if (fetch_ask)
    ask_destroy (fetch_ask);
  fetch_ask = NULL;
}



void
fetch_next (void)
{
  PREAMBLE;

  select_next (fetch_select);
}



void
fetch_prev (void)
{
  PREAMBLE;

  select_prev (fetch_select);
}



void
fetch_next_page (void)
{
  PREAMBLE;

  select_next_page (fetch_select);
}



void
fetch_prev_page (void)
{
  PREAMBLE;

  select_prev_page (fetch_select);
}



void
fetch_first (void)
{
  PREAMBLE;

  select_first (fetch_select);
}



void
fetch_last (void)
{
  PREAMBLE;

  select_last (fetch_select);
}



void
fetch_rset (void)
{
  int i;
  
  PREAMBLE;

  pop_rset ();
  for (i = 0; i < fetch_select->items->count; i++){
    *fetch_select->items->array[i] = ' ';
  }
  select_redraw (fetch_select);
}



void
fetch_del_all (void)
{
  int i;
  
  if (begin_connection ())
    return;

  for (i = 0; i < pop_stat_size; i++){
    pop_dele (i);
  }

  host = NULL;
  
  pop_finish_connection ();
}



void
fetch_del_mail (void)
{
  int index = selected_num ();

  PREAMBLE;

  pop_dele (index);
  *fetch_select->items->array[fetch_select->bar_pos] = 'D';
  select_redraw (fetch_select);
}



void
fetch_get_mail (void)
{
  int index;
  
  PREAMBLE;

  index = selected_index ();

  if (get_mail (index) == 0){
    *fetch_select->items->array[fetch_select->bar_pos] = 'F';
    select_redraw (fetch_select);
    fetch_got_mail = 1;
    fetch_mark_mail ();
  }
}


void
fetch_mark_mail (void)
{
  int index;
  
  PREAMBLE;

  index = selected_index ();
  pop_stat_list[index].fetched = 1;
}



void
fetch_unmark_mail (void)
{
  int index;
  
  PREAMBLE;

  index = selected_index ();
  pop_stat_list[index].fetched = 0;
}



void
fetch_mark_all (void)
{
  int i;

  PREAMBLE;

  for (i = 0; i < pop_stat_size; i++){
    pop_stat_list[i].fetched = 1;
  }
}



void
fetch_unmark_all (void)
{
  int i;

  PREAMBLE;

  for (i = 0; i < pop_stat_size; i++){
    pop_stat_list[i].fetched = 0;
  }
}



void
fetch_save_pop_list (void)
{
  char *dir;
  char *fname;
  FILE *fp;
  int   i;
  
  dir = ask_for_default ("pop_info_dir", NULL);
  if (dir == NULL){
    error_ (0, "no pop3 info dir");
    return;
  }

  fname = file_with_dir (dir, host);
  fp    = fopen (fname, "w");

  if (fp == NULL){
    error_ (errno, "%s", fname);
    return;
  }

  fprintf (fp, "%d\n", pop_stat_size);
  for (i = 0; i < pop_stat_size; i++){
    fprintf (fp, "%d %d %d\n", pop_stat_list[i].num, pop_stat_list[i].size,
             pop_stat_list[i].mark | pop_stat_list[i].fetched);
  }

  fclose (fp);
  xfree (fname);
}



void
fetch_load_pop_list (void)
{
  char *dir;
  char *fname;
  FILE *fp;
  int   i;

  dir = ask_for_default ("pop_info_dir", NULL);
  if (dir == NULL){
    error_ (0, "no pop3 info dir");
    return;
  }

  fname = file_with_dir (dir, host);
  fp    = fopen (fname, "r");

  if (fp == NULL){
    xfree (fname);
    return;
  }

  fscanf (fp, "%d\n", &prev_stat_size);

  prev_stat_list = xmalloc (prev_stat_size * sizeof (pop_stat_t));
  for (i = 0; i < prev_stat_size; i++){
    fscanf (fp, "%d %d %d\n", &prev_stat_list[i].num, &prev_stat_list[i].size,
            &prev_stat_list[i].mark);
  }

  fclose (fp);
  xfree (fname);
}



void
fetch_merge_lists (void)
{
  lcs_t *lcs;

  if (prev_stat_list == NULL)
    return;
  
  lcs = lcs_compute (pop_stat_list,       prev_stat_list,
                     sizeof (pop_stat_t), sizeof (pop_stat_t),
                     pop_stat_size,       prev_stat_size, stats_equal);

  lcs_mark_part (lcs, pop_stat_list,  prev_stat_list,
                 sizeof (pop_stat_t), sizeof (pop_stat_t), mark_stat);
  lcs_destroy (lcs);

  if (prev_stat_list){
    xfree (prev_stat_list);
    prev_stat_list = NULL;
    prev_stat_size = 0;
  }
}



void
fetch_remove_till_limit (void)
{
  char *limit_s;
  int   i;
  int   limit;
  int   maildrop_size = 0;

  if (fetch_ask == NULL)
    return;
    
  limit_s = ask_get_field (fetch_ask, "limit");
  if (limit_s == NULL)
    return;

  limit = strtol (limit_s, &limit_s, 10);
  if (limit_s){
    if (*limit_s == 'K' || *limit_s == 'k')
      limit *= 1024;
    else if (*limit_s == 'M' || *limit_s == 'm')
      limit *= 1024 * 1024;
  }

  for (i = 0; i < pop_stat_size; i++){
    maildrop_size += pop_stat_list[i].size;
  }

  i = 0;
  while (maildrop_size > limit && i < pop_stat_size){
    if (pop_stat_list[i].mark || pop_stat_list[i].fetched){
      if (pop_dele (i) == 0)
        maildrop_size -= pop_stat_list[i].size;
    }
    i++;
  }
}



void
fetch_remove_till_count (void)
{
  char *count_s;
  int   i;
  int   count;
  int   total_count = pop_stat_size;

  if (fetch_ask == NULL)
    return;

  count_s = ask_get_field (fetch_ask, "max_count");
  if (count_s == NULL)
    return;

  count = strtol (count_s, NULL, 10);

  i = 0;
  while (total_count > count && i <= pop_stat_size){
    if (pop_stat_list[i].mark || pop_stat_list[i].fetched){
      if (pop_dele (i) == 0)
        total_count--;
    }
    i++;
  }
}

/****************************************************************************
 *    INTERFACE CLASS BODIES
 ****************************************************************************/
/****************************************************************************
 *
 *    END MODULE fetch.c
 *
 ****************************************************************************/
