/* 
   elmo - ELectronic Mail Operator

   Copyright (C) 2003 rzyjontko

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  

   ----------------------------------------------------------------------

   file handling
   
*/
/****************************************************************************
 *    IMPLEMENTATION HEADERS
 ****************************************************************************/

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <dirent.h>

#include "xmalloc.h"
#include "rstring.h"
#include "file.h"

/****************************************************************************
 *    IMPLEMENTATION PRIVATE DEFINITIONS / ENUMERATIONS / SIMPLE TYPEDEFS
 ****************************************************************************/

#ifndef P_tmpdir
# define P_tmpdir "/tmp"
#endif

/****************************************************************************
 *    IMPLEMENTATION PRIVATE CLASS PROTOTYPES / EXTERNAL CLASS REFERENCES
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE STRUCTURES / UTILITY CLASSES
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION REQUIRED EXTERNAL REFERENCES (AVOID)
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE DATA
 ****************************************************************************/

static size_t  page_size   = 0;
static char   *map_address = NULL;
static off_t   map_offset  = 0;
static size_t  map_length  = 0;

/****************************************************************************
 *    INTERFACE DATA
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE FUNCTION PROTOTYPES
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE FUNCTIONS
 ****************************************************************************/

static void
initialize (void)
{
  page_size = (size_t) sysconf (_SC_PAGESIZE);
}


static void
free_resources (void)
{
  if (map_address)
    munmap (map_address, map_length);
}

/****************************************************************************
 *    INTERFACE FUNCTIONS
 ****************************************************************************/

int
file_part (const char *fname, off_t offset_start, off_t offset_end,
           char **place)
{
  int    fd;
  int    page_offset;
  size_t part_length;

  if (!page_size)
    initialize ();

  page_offset = offset_start % page_size;
  part_length = offset_end - offset_start + 1;
  
  if (map_address)
    munmap (map_address, map_length);

  fd = open (fname, O_RDONLY);
  if (fd == -1)
    return 1;

  map_offset  = offset_start - page_offset;
  map_length  = page_offset  + part_length;
  map_address = mmap (NULL, map_length, PROT_READ | PROT_WRITE,
                      MAP_PRIVATE, fd, map_offset);

  if (map_address == MAP_FAILED)
    return 1;

  map_address[map_length - 1] = '\0';
  *place = map_address + page_offset;

  close (fd);
  return 0;
}



int
file_size (const char *fname)
{
  struct stat st;

  if (stat (fname, &st))
    return -1;
  return st.st_size;
}



int
file_whole (FILE *fp, char **place, size_t *size)
{
  struct stat st;
  int         fd;

  rewind (fp);
  fd = fileno (fp);

  if (fstat (fd, &st))
    return 1;

  *size  = st.st_size;
  *place = mmap (NULL, *size, PROT_READ | PROT_WRITE, MAP_PRIVATE, fd, 0);

  if (*place == MAP_FAILED)
    return 1;
  
  return 0;
}



void
file_free_buffer (char *buffer, size_t size)
{
  munmap (buffer, size);
}



char *
file_with_dir (const char *dir, const char *file)
{
  size_t  dlen   = strlen (dir);
  size_t  flen   = strlen (file);
  char   *result = xmalloc (dlen + 1 + flen + 1);

  memcpy (result, dir, dlen);
  memcpy (result + dlen + 1, file, flen);

  result[dlen] = '/';
  result[dlen + 1 + flen] = '\0';
  
  return result;
}



FILE *
file_temp_file (char **temp_name)
{
  char  name[] = P_tmpdir "/elmo-XXXXXX";
  int   fd;

  while (1){
    fd = mkstemp (name);
    if (fd != -1)
      break;
    if (errno != EEXIST){
      return NULL;
    }
  }
  if (temp_name)
    *temp_name = xstrdup (name);
  return fdopen (fd, "r+");
}



FILE *
file_open (char *fname, const char *mode, int flags)
{
  int   fd;
  char *true_name;
  FILE *fp;

  true_name = file_expand_tilde (fname);
  
  fd = open (true_name, flags, 0644);
  if (fd == -1){
    if (true_name != fname)
      xfree (true_name);
    return NULL;
  }

  fp = fdopen (fd, mode);

  if (true_name != fname)
    xfree (true_name);
  return fp;
}



void
file_free_resources (void)
{
  free_resources ();
}




int
file_rename (const char *oldname, const char *newname)
{
  struct stat st;
  int         ofd;
  int         nfd;
  char       *ofile;
  char       *nfile;
  
  if (rename (oldname, newname) == 0)
    return 0;

  if (errno != EXDEV)
    return 1;

  /**
   * step 0: open old file
   */
  ofd = open (oldname, O_RDONLY);
  if (ofd == -1)
    return 1;

  /**
   * step 1: get old file size
   */
  if (fstat (ofd, &st)){
    close (ofd);
    return 1;
  }
  
  /**
   * step 2: open new file for writing
   */
  nfd = open (newname, O_RDWR | O_CREAT | O_EXCL, 0600);
  if (nfd == -1){
    close (ofd);
    return 1;
  }

  /**
   * step 3: resize new file
   */
  if (ftruncate (nfd, st.st_size)){
    close (ofd);
    close (nfd);
    return 1;
  }

  /**
   * step 4: mmap both files
   */
  ofile = mmap (NULL, st.st_size, PROT_READ, MAP_PRIVATE, ofd, 0);
  if (ofile == MAP_FAILED){
    close (ofd);
    close (nfd);
    return 1;
  }
  nfile = mmap (NULL, st.st_size, PROT_READ | PROT_WRITE, MAP_SHARED, nfd, 0);
  if (nfile == MAP_FAILED){
    close (ofd);
    close (nfd);
    return 1;
  }

  /**
   * step 5: copy file contents
   */
  memcpy (nfile, ofile, st.st_size);

  /**
   * step 6: unmap regions and close file descriptors
   */
  munmap (ofile, st.st_size);
  munmap (nfile, st.st_size);
  close (ofd);
  close (nfd);

  return 0;
}



char *
file_expand_tilde (char *dir)
{
  char *result;
  char *home = getenv ("HOME");
  int   hlen = (home) ? strlen (home) : 0;
  int   dlen = (dir) ? strlen (dir) : 0;
  
  if (dir == NULL)
    return NULL;
  
  if (*dir != '~')
    return dir;

  result = xmalloc (hlen + dlen);
  memcpy (result, home, hlen);
  memcpy (result + hlen, dir + 1, dlen);
  return result;
}



rstring_t *
file_dir_items (const char *dir, int (*select)(const struct dirent *))
{
#ifdef HAVE_SCANDIR
  int             i;
  int             len;
  int             count;
  struct stat     st;
  struct dirent **list;
  rstring_t      *result;
  char           *file;

  if (dir == NULL)
    return NULL;

  count = scandir (dir, &list, select, alphasort);
  if (count == -1){
    return NULL;
  }

  result = rstring_create_size (count + 1);
  for (i = 0; i < count; i++){
    file = file_with_dir (dir, list[i]->d_name);
    if (stat (file, &st)){
      xfree (file);
      continue;
    }
    xfree (file);
    
    if (S_ISDIR (st.st_mode)){
      len  = strlen (list[i]->d_name);
      file = xmalloc (len + 2);
      memcpy (file, list[i]->d_name, len);
      file[len]     = '/';
      file[len + 1] = '\0';
    }
    else {
      file = xstrdup (list[i]->d_name);
    }
    rstring_add (result, file);
  }
  result->allocated_all = 1;

  for (i = 0; i < count; i++)
    free (list[i]);

  if (count)
    free (list);

  return result;

#else
  return NULL;
#endif
}

/****************************************************************************
 *    INTERFACE CLASS BODIES
 ****************************************************************************/
/****************************************************************************
 *
 *    END MODULE file.c
 *
 ****************************************************************************/
