/* 
   elmo - ELectronic Mail Operator

   Copyright (C) 2002, 2003 rzyjontko
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  

   ----------------------------------------------------------------------

   handle mail folder
   
*/
/****************************************************************************
 *    IMPLEMENTATION HEADERS
 ****************************************************************************/

#include <string.h>

#include "ecurses.h"
#include "folder.h"
#include "xmalloc.h"
#include "line.h"
#include "screen.h"
#include "status.h"
#include "error.h"
#include "ask.h"
#include "cmd.h"
#include "wrapbox.h"
#include "bayes.h"
#include "progress.h"
#include "mybox.h"
#include "hook.h"
#include "exec.h"
#include "eprintf.h"
#include "read.h"
#include "maildir.h"

/****************************************************************************
 *    IMPLEMENTATION PRIVATE DEFINITIONS / ENUMERATIONS / SIMPLE TYPEDEFS
 ****************************************************************************/

#define MIN(a,b) (((a)<(b))?(a):(b))


#define PREAMBLE do { if (wrapbox_mail_count () < 0) return; } while (0)

/****************************************************************************
 *    IMPLEMENTATION PRIVATE CLASS PROTOTYPES / EXTERNAL CLASS REFERENCES
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE STRUCTURES / UTILITY CLASSES
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION REQUIRED EXTERNAL REFERENCES (AVOID)
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE DATA
 ****************************************************************************/

/**
 * There may be few pages filled with list of mails.  pages holds array
 * of WINDOW pointers.  current_page_index holds page which should be on
 * the top of visibility.  current_page_index is recalculated in set_bar.
 */
static WINDOW **pages              = NULL;
static int      current_page_index = 0;

/**
 * page_size holds number of lines in single WINDOW object
 * page_count holds size of pages array
 */
static int page_size  = 0;
static int page_count = 0;

/**
 * Bar is a hilighted position.  bar holds this window (which is then
 * moved) and bar_index holds index which is currently chosen.
 */
static WINDOW *bar       = NULL;
static int     bar_index = 0;

/**
 * This is used in print_line.
 */
static char *line_txt = NULL;


/****************************************************************************
 *    INTERFACE DATA
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE FUNCTION PROTOTYPES
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE FUNCTIONS
 ****************************************************************************/


/**
 * these 2 functions are used when printing a tree
 * tree itself is stored without colour, but it is coloured at display time
 * it's because bar has to be printed in different colour
 */
static void
set_color (chtype *str, int color)
{
  while (*str){
    *str |= color;
    str++;
  }
}



static void
unset_color (chtype *str, int color)
{
  while (*str){
    *str &= ~color;
    str++;
  }
}



static void
print_line (int index, WINDOW *win, int y, int x, int color)
{
  char        *txt;
  mail_t      *mail = mail_array_get (wrapbox_marray, index);

  if (mail == NULL)
    return;
  
  if (line_txt == NULL)
    line_txt = xmalloc (COLS + 2);
  
  txt = eprintf_mail (mybox_format, mail, line_txt);
  
  mvwaddstr (win, y, x, txt);

  if (wrapbox_marray->order == ORDER_THREAD && mail->tree && mail->tree_len){
    set_color (mail->tree, color);
    waddchstr (win, mail->tree);
    unset_color (mail->tree, color);
    getyx (win, y, x);
    wmove (win, y, x + mail->tree_len - 1);
  }

  getyx (win, y, x);
  if (mail->subject)
    waddnstr (win, mail->subject, COLS - x - 1);

  wclrtoeol (win);
}


static void
copy_tree (chtype *to, chtype *from)
{
  while (*from && *from != '>'){
    if (*from == ' ' || *from == ACS_VLINE)
      *to = *from;
    else if (*from == ACS_LLCORNER || *from == ACS_HLINE)
      *to = ' ';
    else if (*from == ACS_LTEE)
      *to = ACS_VLINE;
    else
      *to = *from;
    to++;
    from++;
  }
}


static void
make_trees (void)
{
  int     i;
  int     mail_count;
  mail_t *mail;
  mail_t *parent;

  mail_count = wrapbox_mail_count ();
  
  for (i = 0; i < mail_count; i++){
    mail   = mail_array_get (wrapbox_marray, i);
    parent = mail_array_get (wrapbox_marray, mail->parent);
    if (mail->tree)
      xfree (mail->tree);
    if (parent){
      mail->tree_len = parent->tree_len + ((parent->tree_len) ? 2 : 4);
      mail->tree     = xmalloc (mail->tree_len * sizeof (chtype));
      mail->tree[mail->tree_len - 3] = ACS_HLINE;
      mail->tree[mail->tree_len - 2] = '>';
      mail->tree[mail->tree_len - 1] = 0;
      copy_tree (mail->tree, parent->tree);
      if (mail->is_last_child)
        mail->tree[mail->tree_len - 4] = ACS_LLCORNER;
      else
        mail->tree[mail->tree_len - 4] = ACS_LTEE;
    }
    else {
      mail->tree_len = 0;
      mail->tree     = xcalloc (1, sizeof (chtype));
    }
  }
}



static void
free_resources (void)
{
  int i;

  if (line_txt){
    xfree (line_txt);
    line_txt = NULL;
  }
  
  if (pages == NULL)
    return;

  for (i = 0; i < page_count; i++)
    delwin (pages[i]);
  xfree (pages);

  pages      = NULL;
  page_count = 0;

  if (bar)
    delwin (bar);
  bar = NULL;
}



static void
update_bar (void)
{
  if (page_size == 0)
    return;
  
  current_page_index = bar_index / page_size;
  
  if (!bar){
    bar = newwin (1, COLS, bar_index % page_size, 0);
    wbkgdset (bar, COLOR_PAIR (2));
  }
  else {
    mvwin (bar, bar_index % page_size, 0);
  }
}



static void
set_bar (void)
{
  update_bar ();
  print_line (bar_index, bar, 0, 0, COLOR_PAIR (3));
}




static void
set_window_succession (void)
{
  screen_lock ();
  set_bar ();
  
  if (pages){
    touchwin (pages[current_page_index]);
    wnoutrefresh (pages[current_page_index]);
    touchwin (bar);
    wnoutrefresh (bar);
  }
  screen_unlock ();
}




static void
fill_page (int num)
{
  WINDOW *page  = pages[num];
  int     index = num * page_size;
  int     i;

  wclear (page);
  for (i = 0; i < page_size; index++, i++){
    if (index >= wrapbox_marray->count)
      break;
    print_line (index, page, i, 0, COLOR_PAIR (4));
  }
}



static void
put_mails_on_pages (void)
{
  int i;

  if (pages)
    free_resources ();
  
  page_size  = LINES - 2;
  page_count = wrapbox_marray->count / page_size + 1;

  pages = xmalloc (page_count * sizeof (WINDOW *));

  screen_lock ();
  for (i = 0; i < page_count; i++){
    pages[i] = newwin (page_size, COLS, 0, 0);
    fill_page (i);
    touchwin (pages[i]);
    wnoutrefresh (pages[i]);
  }
  screen_unlock ();
}




static void
display_status_info (void)
{
  int   count;
  int   read;
  int   old;
  int   new;
  char *seek = NULL;

  if (wrapbox_marray == NULL)
    return;
  
  if (ask_for_default ("relative_names", NULL))
    seek = strrchr (wrapbox_marray->path, '/');

  if (seek == NULL)
    seek = wrapbox_marray->path;
  else
    seek++;
  
  count = wrapbox_marray->count;
  read  = mail_array_read (wrapbox_marray);
  old   = mail_array_old (wrapbox_marray);
  new   = count - read - old;
  
  status_put_mailbox (seek, count, old, new);
}


/****************************************************************************
 *    INTERFACE FUNCTIONS
 ****************************************************************************/

mail_t *
folder_mail_selected (void)
{
  int     mail_count = wrapbox_mail_count ();
  mail_t *mail;
  
  if (mail_count == -1 || bar_index < 0 || bar_index >= mail_count)
    return NULL;

  mail = mail_array_get (wrapbox_marray, bar_index);
  return mail;
}



void
folder_init (void)
{
  char *box;

  box = mybox_start ();
  
  mybox_switching_to (box);
  folder_read_box (box);
  xfree (box);
}




void
folder_free_resources (void)
{
  folder_at_leave ();
  free_resources ();
}



int
folder_read_box (char *name)
{
  if (wrapbox_read_box (name)){
    error_ (0, "%s is not a valid mailbox", name);
    return 1;
  }
  
  display_status_info ();
  free_resources ();
  mail_array_sort_date (wrapbox_marray);

  bar_index = mail_array_first_unread_index (wrapbox_marray);
  if (bar_index == -1)
    bar_index = wrapbox_mail_count () - 1;

  folder_spam_check_all ();
  folder_sort_threads ();

  if (wrapbox_mail_count () == 0){
    put_mails_on_pages ();
    folder_refresh ();
  }
  return 0;
}


void
folder_soft_refresh (void)
{
  set_window_succession ();
  display_status_info ();
}


void
folder_refresh (void)
{
  int i;
  
  if (pages){
    screen_lock ();
    for (i = 0; i < page_count; i++){
      fill_page (i);
      wnoutrefresh (pages[i]);
    }
    screen_unlock ();
    set_window_succession ();
    screen_redrawwin (pages[current_page_index]);
    if (bar)
      screen_redrawwin (bar);
    display_status_info ();
  }
}



void
folder_update (void)
{
  mail_t *mail;
  void   *ptr;
  void   *ptr_next;
  
  PREAMBLE;
  
  mail = mail_array_get (wrapbox_marray, bar_index);
  ptr  = (mail) ? mail->place.generic : NULL;
  
  mail = mail_array_get (wrapbox_marray, bar_index + 1);
  if (mail == NULL)
    mail = mail_array_get (wrapbox_marray, bar_index);
  ptr_next = (mail) ? mail->place.generic : NULL;
  
  wrapbox_refresh ();
  free_resources ();
  mail_array_sort_date (wrapbox_marray);

  bar_index = mail_array_find_place (wrapbox_marray, ptr);
  if (bar_index == -1)
    bar_index = mail_array_find_place (wrapbox_marray, ptr_next);

  folder_spam_check_all ();
  folder_sort_threads ();

  display_status_info ();
}



void
folder_flush (void)
{
  PREAMBLE;
  
  folder_spam_check_all ();
  folder_spam_flush ();
  folder_update ();
}



void
folder_redraw_bar (void)
{
  PREAMBLE;
  
  if (cmd_state_get (0) == CMD_LIST){
    print_line (bar_index, pages[current_page_index],
                bar_index % page_size, 0, COLOR_PAIR (4));
    set_window_succession ();
  }
}



/****************************************************************************
 *    SORTING FUNCTIONS
 ****************************************************************************/



void
folder_sort_threads (void)
{
  mail_t *mail;
  void   *ptr = NULL;
  
  PREAMBLE;
  
  mail = mail_array_get (wrapbox_marray, bar_index);

  if (mail){
    ptr = mail->place.generic;
  }
  
  mail_array_sort_threads (wrapbox_marray);

  if (mail){
    bar_index = mail_array_find_place (wrapbox_marray, ptr);
  }
  else {
    bar_index = wrapbox_marray->count - 1;
  }

  make_trees ();
  put_mails_on_pages ();
  set_window_succession ();
}



void
folder_sort_date (void)
{
  mail_t *mail;
  void   *ptr = NULL;
  
  PREAMBLE;
  
  mail = mail_array_get (wrapbox_marray, bar_index);
  
  if (mail) {
    ptr  = mail->place.generic;
  }
  
  mail_array_sort_date (wrapbox_marray);

  bar_index = mail_array_find_place (wrapbox_marray, ptr);

  put_mails_on_pages ();
  set_window_succession ();
}



void
folder_sort_from (void)
{
  mail_t *mail;
  void   *ptr = NULL;
  
  PREAMBLE;
  
  mail = mail_array_get (wrapbox_marray, bar_index);
  
  if (mail) {
    ptr  = mail->place.generic;
  }
  
  mail_array_sort_from (wrapbox_marray);

  bar_index = mail_array_find_place (wrapbox_marray, ptr);

  put_mails_on_pages ();
  set_window_succession ();
}



void
folder_sort_subject (void)
{
  mail_t *mail;
  void   *ptr = NULL;
  
  PREAMBLE;
 
  mail = mail_array_get (wrapbox_marray, bar_index);
  
  if (mail) {
    ptr  = mail->place.generic;
  }
  
  mail_array_sort_subject (wrapbox_marray);

  bar_index = mail_array_find_place (wrapbox_marray, ptr);

  put_mails_on_pages ();
  set_window_succession ();
}


/****************************************************************************
 *    BAR MOVEMENT FUNCTIONS
 ****************************************************************************/


void
folder_page_next (void)
{
  PREAMBLE;
  
  if (current_page_index == page_count - 1)
    bar_index = wrapbox_marray->count - 1;
  else
    bar_index = MIN (bar_index + page_size, wrapbox_marray->count - 1);
  set_window_succession ();
}



void
folder_page_prev (void)
{
  PREAMBLE;
  
  if (current_page_index == 0)
    bar_index = 0;
  else
    bar_index -= page_size;
  set_window_succession ();
}



void
folder_bar_next (void)
{
  PREAMBLE;
  
  if (bar_index == wrapbox_marray->count - 1)
    return;
  bar_index++;
  set_window_succession ();
}




void
folder_bar_prev (void)
{
  PREAMBLE;
  
  if (bar_index == 0)
    return;
  bar_index--;
  set_window_succession ();
}



void
folder_bar_first (void)
{
  PREAMBLE;
  
  bar_index = 0;
  set_window_succession ();
}



void
folder_bar_last (void)
{
  PREAMBLE;
  
  bar_index = wrapbox_marray->count - 1;
  set_window_succession ();
}



/****************************************************************************
 *    ACTIONS PRIVATE FUNCTIONS
 ****************************************************************************/

static void
action_move_mail (mail_t *mail, char *box)
{
  wrapbox_move_mail_to (mail, box);
}



static void
action_make_read (mail_t *mail, char *nothing)
{
  if (mail == NULL)
    return;

  mail->flags |= FLAG_READ;
}


static void
action_check_spam (mail_t *mail, char *nothing)
{
  if (bayes_is_spam (mail)){
    mail->flags &= ~ FLAG_NOSPAM;
    mail->flags &= ~ FLAG_LEGITIMATE;
    mail->flags |= FLAG_SPAM;
  }
  else {
    mail->flags &= ~ FLAG_SPAM;
    mail->flags |= FLAG_NOSPAM;
  }
}


static void
action_delete_spam (mail_t *mail, char *trash)
{
  if (mail->flags & FLAG_LEGITIMATE){
    bayes_unscan_legitimate (mail);
  }

  bayes_scan_spam (mail);
  wrapbox_move_mail_to (mail, trash);
}



static void
action_is_not_spam (mail_t *mail, char *nothing)
{
  mail->flags |= FLAG_NOSPAM;
  mail->flags &= ~ FLAG_SPAM;
}



static void
action_remove_flag (mail_t *mail, char *flag)
{
  if (*flag == 'L')
    mail->flags &= ~ FLAG_LEGITIMATE;
}


static void
take_action (void (*fun)(mail_t *, char*), char *arg)
{
  int     i;
  int     took = 0;
  mail_t *mail;

  PREAMBLE;
  
  for (i = 0; i < wrapbox_marray->count; i++){
    mail = mail_array_get (wrapbox_marray, i);
    if (mail->flags & FLAG_FLAGGED){
      mail->flags &= ~ FLAG_FLAGGED;
      fun (mail, arg);
      took = 1;
    }
  }

  if (! took){
    mail = mail_array_get (wrapbox_marray, bar_index);
    fun (mail, arg);
  }
}

/****************************************************************************
 *    ACTIONS INTERFACE FUNCTIONS
 ****************************************************************************/

void
folder_after_read (void)
{
  mail_t *mail;
  
  PREAMBLE;

  mail = mail_array_get (wrapbox_marray, bar_index);
  
  if (mail == NULL)
    return;
  
  mail->flags |= FLAG_OLD;
  mail->flags |= FLAG_READ;

  wrapbox_marray->old_count  = -1;
  wrapbox_marray->read_count = -1;
  
  folder_redraw_bar ();
  display_status_info ();
}


void
folder_after_reply (void)
{
  mail_t *mail;
  
  PREAMBLE;

  mail         = mail_array_get (wrapbox_marray, bar_index);
  
  if (mail == NULL)
    return;
  
  mail->flags |= FLAG_ANSWERED;

  folder_after_read ();
}



void
folder_after_fwd (void)
{
  mail_t *mail;
  
  PREAMBLE;

  mail         = mail_array_get (wrapbox_marray, bar_index);
  
  if (mail == NULL)
    return;
  
  mail->flags |= FLAG_PASSED;

  folder_after_read ();
}



void
folder_delete_mail (void)
{
  char *box;

  PREAMBLE;
  
  box = mybox_trash ();

  take_action (action_move_mail, box);
  xfree (box);

  folder_update ();
}

void
folder_delete_mail_permanently (void)
{
  mail_t *mail;

  PREAMBLE;
  
  mail = mail_array_get (wrapbox_marray, bar_index);
  
  if (mail == NULL)
    return;

  maildir_remove (mail);
  folder_update ();
}

void
folder_move_mail (void)
{
  char *box;

  PREAMBLE;
  
  box = mybox_select_subdir ();

  if (box == NULL)
    return;

  take_action (action_move_mail, box);
  xfree (box);

  folder_update ();
}



void
folder_make_read (void)
{
  PREAMBLE;
  
  take_action (action_make_read, NULL);
  folder_refresh ();
}



void
folder_at_leave (void)
{
  exec_t *exec;

  exec = exec_lookup_fun (folder_at_leave);

  hook_execute (exec->hook);

  folder_spam_flush ();
  wrapbox_apply_flags ();
  wrapbox_dump_box ();
}


/****************************************************************************
 *    FLAGGING FUNCTIONS
 ****************************************************************************/


void
folder_toggle_flag (void)
{
  mail_t *mail;
  
  PREAMBLE;

  mail = mail_array_get (wrapbox_marray, bar_index);
  
  if (mail == NULL)
    return;

  if (mail->flags & FLAG_FLAGGED)
    mail->flags &= ~ FLAG_FLAGGED;
  else
    mail->flags |= FLAG_FLAGGED;

  folder_redraw_bar ();
}



void
folder_flag_all (void)
{
  int     i;
  mail_t *mail;

  PREAMBLE;
  
  for (i = 0; i < wrapbox_marray->count; i++){
    mail = mail_array_get (wrapbox_marray, i);
    mail->flags |= FLAG_FLAGGED;
  }

  folder_refresh ();
}



void
folder_unflag_all (void)
{
  int     i;
  mail_t *mail;

  PREAMBLE;
  
  for (i = 0; i < wrapbox_marray->count; i++){
    mail = mail_array_get (wrapbox_marray, i);
    mail->flags &= ~ FLAG_FLAGGED;
  }

  folder_refresh ();
}




void
folder_flag_invert (void)
{
  int     i;
  mail_t *mail;

  PREAMBLE;
  
  for (i = 0; i < wrapbox_marray->count; i++){
    mail = mail_array_get (wrapbox_marray, i);
    if (mail->flags & FLAG_FLAGGED)
      mail->flags &= ~ FLAG_FLAGGED;
    else
      mail->flags |= FLAG_FLAGGED;
  }

  folder_refresh ();
}


void
folder_flag_duplicates (void)
{
  int     i;
  mail_t *mail;
  
  PREAMBLE;

  for (i = 0; i < wrapbox_marray->count; i++){
    mail = mail_array_get (wrapbox_marray, i);
    if (mail->flags & FLAG_DUPLICATE)
      mail->flags |= FLAG_FLAGGED;
  }

  folder_refresh ();
}



void
folder_remove_flag (void)
{
  char *flag;

  PREAMBLE;
  
  flag = read_argument ("flag: ", NULL, COMPLETE_FILES);

  if (flag && *flag != '\0')
    take_action (action_remove_flag, flag);
  folder_update ();
}



/****************************************************************************
 *    SPAM FUNCTIONS
 ****************************************************************************/

void
folder_spam_check (void)
{
  if (! mybox_protect)
    return;
  
  PREAMBLE;
  
  take_action (action_check_spam, NULL);
  folder_refresh ();
}



void
folder_spam_check_all (void)
{
  int     i;
  mail_t *mail;

  if (! mybox_protect)
    return;
  
  PREAMBLE;
  
  for (i = 0; i < wrapbox_marray->count; i++){
    mail = mail_array_get (wrapbox_marray, i);
    if ((mail->flags & FLAG_SPAM) || (mail->flags & FLAG_NOSPAM)
        || (mail->flags & FLAG_LEGITIMATE))
      continue;
    if (bayes_is_spam (mail))
      mail->flags |= FLAG_SPAM;
    else
      mail->flags |= FLAG_NOSPAM;
  }
  folder_refresh ();
}



void
folder_spam_delete (void)
{
  char *box;

  if (! mybox_protect)
    return;
  
  PREAMBLE;
  
  box = mybox_spam ();

  take_action (action_delete_spam, box);
  xfree (box);

  folder_update ();
}



void
folder_spam_is_not (void)
{
  if (! mybox_protect)
    return;
  
  PREAMBLE;
  
  take_action (action_is_not_spam, NULL);
  folder_refresh ();
}



void
folder_spam_flush (void)
{
  int     i;
  char   *box;
  mail_t *mail;

  if (! mybox_protect)
    return;

  PREAMBLE;
  
  box = mybox_spam ();
  
  for (i = 0; i < wrapbox_marray->count; i++){
    mail = mail_array_get (wrapbox_marray, i);
    if (mail->flags & FLAG_SPAM){
      bayes_scan_spam (mail);
      wrapbox_move_mail_to (mail, box);
    }
    else if ((mail->flags & FLAG_NOSPAM)
             && ! (mail->flags & FLAG_LEGITIMATE)){
      mail->flags |= FLAG_LEGITIMATE;
      bayes_scan_legitimate (mail);
    }
  }

  if (box)
    xfree (box);
}

/****************************************************************************
 *    INTERFACE CLASS BODIES
 ****************************************************************************/
/****************************************************************************
 *
 *    END MODULE folder.c
 *
 ****************************************************************************/
