/* 
   elmo - ELectronic Mail Operator

   Copyright (C) 2003 rzyjontko

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  

   ----------------------------------------------------------------------

   chtype strings operations
   
*/
/****************************************************************************
 *    IMPLEMENTATION HEADERS
 ****************************************************************************/

#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <regex.h>

#include "ecurses.h"
#include "xmalloc.h"
#include "rarray.h"
#include "rchtype.h"
#include "rstring.h"
#include "error.h"

/****************************************************************************
 *    IMPLEMENTATION PRIVATE DEFINITIONS / ENUMERATIONS / SIMPLE TYPEDEFS
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE CLASS PROTOTYPES / EXTERNAL CLASS REFERENCES
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE STRUCTURES / UTILITY CLASSES
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION REQUIRED EXTERNAL REFERENCES (AVOID)
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE DATA
 ****************************************************************************/

static int mask = 0;

static regmatch_t matches[1];

static regmatch_t b_matches[1];
static regmatch_t u_matches[1];

/****************************************************************************
 *    INTERFACE DATA
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE FUNCTION PROTOTYPES
 ****************************************************************************/

static void    copy_string (chtype *dest, char *str);
static chtype *search_line_end (char **txt, size_t len);

/****************************************************************************
 *    IMPLEMENTATION PRIVATE FUNCTIONS
 ****************************************************************************/

static void
copy_string (chtype *dest, char *str)
{
  char *tmp;

  for (tmp = str; *tmp; tmp++){

    if (tmp - str == b_matches[0].rm_so)
      mask |= A_BOLD;
    if (tmp - str == b_matches[0].rm_eo)
      mask &= ~ A_BOLD;

    if (tmp - str == u_matches[0].rm_so)
      mask |= A_UNDERLINE;
    if (tmp - str == u_matches[0].rm_eo)
      mask &= ~ A_UNDERLINE;
    
    *dest = (*tmp & ~A_ATTRIBUTES) | mask;
    dest++;
  }

  if (tmp - str == b_matches[0].rm_eo)
    mask &= ~ A_BOLD;
  if (tmp - str == u_matches[0].rm_eo)
    mask &= ~ A_UNDERLINE;

  *dest = '\0';
}


static chtype *
chstrdup (char *txt, int len)
{
  chtype *result = xmalloc ((len + 1) * sizeof (chtype));
  copy_string (result, txt);
  return result;
}


static chtype *
search_line_end (char **txt, size_t line_len)
{
  char    oldchar;
  char   *seek;
  char   *last_space = NULL;
  chtype *result;

  for (seek = *txt; line_len && *seek; seek++){

    switch (*seek){

      case '\t': case ' ':
        last_space = seek;
        line_len--;
        break;

      case '\r':
        if (seek[1] == '\n'){
          *seek  = '\0';
          result = chstrdup (*txt, seek - *txt);
          *seek  = '\r';
          *txt   = seek + 2;
          return result;
        }
        else if (seek[1] == '\0'){
          *seek  = '\0';
          result = chstrdup (*txt, seek - *txt);
          *seek  = '\r';
          *txt   = NULL;
          return result;
        }
        break;
      
      case '\n':
        *seek  = '\0';
        result = chstrdup (*txt, seek - *txt);
        *seek  = '\n';
        *txt   = seek + 1;
        return result;

      default:
        line_len--;
        break;
    }
  }

  if (*seek == '\0'){
    result = chstrdup (*txt, seek - *txt);
    *txt   = NULL;
    return result;
  }

  if (last_space){
    oldchar     = *last_space;
    *last_space = '\0';
    result      = chstrdup (*txt, last_space - *txt);
    *last_space = oldchar;
    *txt        = last_space + 1;
    return result;
  }

  oldchar = *seek;
  *seek   = '\0';
  result  = chstrdup (*txt, seek - *txt);
  *seek   = oldchar;
  *txt    = seek;
  return result;
}


static void
add_line (rchtype_t *items, char *line, size_t line_len,
          regex_t *b_re, regex_t *u_re)
{
  int     ret;
  chtype *item;
  
  while (1){

    ret = regexec (b_re, line, 1, b_matches, 0);
    if (ret && ret != REG_NOMATCH){
      error_regex (ret, b_re);
      regfree (b_re);
      return;
    }
    else if (ret == REG_NOMATCH){
      b_matches[0].rm_so = -1;
      b_matches[0].rm_eo = -1;
    }

    ret = regexec (u_re, line, 1, u_matches, 0);
    if (ret && ret != REG_NOMATCH){
      error_regex (ret, u_re);
      regfree (u_re);
      return;
    }
    else if (ret == REG_NOMATCH){
      u_matches[0].rm_so = -1;
      u_matches[0].rm_eo = -1;
    }

    item = search_line_end (&line, line_len);
    rchtype_add (items, item);
    if (line == NULL)
      break;
  }
}



static int
regex_init (regex_t *compiled, char *re)
{
  int ret;
  
  ret = regcomp (compiled, re, REG_ICASE | REG_EXTENDED);

  if (ret){
    error_regex (ret, compiled);
    regfree (compiled);
    return 1;
  }
  return 0;
}

/****************************************************************************
 *    INTERFACE FUNCTIONS
 ****************************************************************************/

rchtype_t *
rchtype_create_size (int size)
{
  return (rchtype_t *) rarray_create_size (size);
}



rchtype_t *
rchtype_create (void)
{
  return (rchtype_t *) rarray_create ();
}



void
rchtype_destroy (rchtype_t *ptr)
{
  chtype **str;

  for (str = ptr->array; *str; str++){
    xfree (*str);
  }
  rarray_destroy ((rarray_t *) ptr);
}




void
rchtype_add (rchtype_t *ptr, chtype *txt)
{
  rarray_add ((rarray_t *) ptr, txt);
}



void
rchtype_add_string (rchtype_t *ptr, char *fmt, ...)
{
  va_list      ap;
  chtype      *txt;
  static char  str[100];
  size_t       len;

  va_start (ap, fmt);
  len = vsnprintf (str, sizeof (str), fmt, ap);
  va_end (ap);

  if (len == -1){
    len      = sizeof (str) - 1;
    str[len] = '\0';
  }
  
  txt = xmalloc ((len + 1) * sizeof (chtype));
  copy_string (txt, str);
  rchtype_add (ptr, txt);
}



void
rchtype_remove (rchtype_t *ptr, unsigned index)
{
  rarray_remove ((rarray_t *) ptr, index);
}



void
rchtype_shrink (rchtype_t *ptr)
{
  rarray_shrink ((rarray_t *) ptr);
}



void
rchtype_set_mask (int modifier)
{
  mask |= modifier;
}


void
rchtype_unset_mask (int modifier)
{
  mask &= ~modifier;
}


rchtype_t *
rchtype_split_lines (char *str, size_t line_len)
{
  regex_t    compiled;
  regex_t    bold;
  regex_t    underline;
  regex_t    dashes;
  rstring_t *lines;
  rchtype_t *result;
  int        i;
  int        ret;

  if (regex_init (&compiled, "^[[:blank:]]*[[:alnum:]]{0,4}(>)"))
    return NULL;

  if (regex_init (&dashes, "^\\-\\-[[:blank:]]*$"))
    return NULL;
  
  if (regex_init (&bold, "\\*[[:alnum:]]([^*]*[[:alnum:]])?\\*"))
    return NULL;
  
  if (regex_init (&underline, "_[[:alnum:]](.*[[:alnum:]])?_"))
    return NULL;
  
  lines  = rstring_split (str, "\n");
  result = rchtype_create ();
  mask   = 0;

  for (i = 0; i < lines->count; i++){

    ret = regexec (&dashes, lines->array[i], 1, matches, 0);

    if (ret && ret != REG_NOMATCH){
      error_regex (ret, &dashes);
      rchtype_destroy (result);
      regfree (&dashes);
      return NULL;
    }
    else if (ret == 0){
      mask = COLOR_PAIR (5);
    }

    ret = regexec (&compiled, lines->array[i], 1, matches, 0);

    if (ret && ret != REG_NOMATCH){
      error_regex (ret, &compiled);
      rchtype_destroy (result);
      regfree (&compiled);
      return NULL;
    }
    else if (ret == REG_NOMATCH){
      if (mask != COLOR_PAIR (5))
        mask = 0;
    }
    else {
      mask = COLOR_PAIR (6);
    }

    add_line (result, lines->array[i], line_len, &bold, &underline);
  }

  mask = 0;
  
  rstring_delete (lines);
  regfree (&compiled);
  regfree (&bold);
  regfree (&underline);

  return result;
}



rchtype_t *
rchtype_join (rchtype_t *x, rchtype_t *y)
{
  return (rchtype_t *) rarray_join ((rarray_t *) x, (rarray_t *) y);
}


/****************************************************************************
 *    INTERFACE CLASS BODIES
 ****************************************************************************/
/****************************************************************************
 *
 *    END MODULE rchtype.c
 *
 ****************************************************************************/
