/* 
   elmo - ELectronic Mail Operator

   Copyright (C) 2003 rzyjontko

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  

   ----------------------------------------------------------------------

   
*/
/****************************************************************************
 *    IMPLEMENTATION HEADERS
 ****************************************************************************/

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <string.h>
#include <dirent.h>
#include <ctype.h>

#include "ecurses.h"
#include "screen.h"
#include "clock.h"
#include "xmalloc.h"
#include "cmd.h"
#include "choose.h"
#include "rstring.h"
#include "read.h"
#include "file.h"
#include "abook.h"

/****************************************************************************
 *    IMPLEMENTATION PRIVATE DEFINITIONS / ENUMERATIONS / SIMPLE TYPEDEFS
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE CLASS PROTOTYPES / EXTERNAL CLASS REFERENCES
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE STRUCTURES / UTILITY CLASSES
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION REQUIRED EXTERNAL REFERENCES (AVOID)
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE DATA
 ****************************************************************************/

static complete_mode_t mode = COMPLETE_NONE;

static int prompt_len = 0;
static int cancel     = 0;
static int tab        = 0;
static int bufsize    = 0;

static char *buffer   = NULL;
static char *position = NULL;

static char      *compl_string;
static rstring_t *compl_items;

/****************************************************************************
 *    INTERFACE DATA
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE FUNCTION PROTOTYPES
 ****************************************************************************/
/****************************************************************************
 *    CURSOR MOTION FUNCTIONS
 ****************************************************************************/

static void
next_char (void)
{
  int y, x;
  
  if (*position == '\0')
    return;
  position++;
  getyx (cmd_echo_area, y, x);
  wmove (cmd_echo_area, y, x + 1);
}


static void
prev_char (void)
{
  int y, x;
  
  if (position == buffer)
    return;
  position--;
  getyx (cmd_echo_area, y, x);
  wmove (cmd_echo_area, y, x - 1);
}


static void
first_char (void)
{
  position = buffer;
  wmove (cmd_echo_area, 0, prompt_len);
}



static void
last_char (void)
{
  int y, x;

  getyx (cmd_echo_area, y, x);
  while (*position != '\0'){
    x++;
    position++;
  }
  wmove (cmd_echo_area, y, x);
}

/****************************************************************************
 *    EDITING FUNCTIONS
 ****************************************************************************/


static void
backspace (void)
{
  int y, x;
  int len;

  if (position == buffer)
    return;
  
  len = strlen (position);
  position--;
  memmove (position, position + 1, len + 1);
  getyx (cmd_echo_area, y, x);
  wmove (cmd_echo_area, y, x - 1);
  wclrtoeol (cmd_echo_area);
  mvwaddstr (cmd_echo_area, y, x - 1, position);
  wmove (cmd_echo_area, y, x - 1);
}



static void
kill_char (void)
{
  if (*position == '\0')
    return;
  next_char ();
  backspace ();
}



static void
kill_line (void)
{
  *position = '\0';
  wclrtoeol (cmd_echo_area);
}


static void
echo_char (int c)
{
  int y, x;
  int len;
  
  if (*position == '\0'){
    waddch (cmd_echo_area, c);
    *position = c;
    position++;
    *position = '\0';
  }
  else {
    len = strlen (position);
    memmove (position + 1, position, len + 1);
    *position = c;
    position++;

    getyx (cmd_echo_area, y, x);
    wclrtoeol (cmd_echo_area);
    waddstr (cmd_echo_area, position - 1);
    wmove (cmd_echo_area, y, x + 1);
  }
}


/****************************************************************************
 *    COMPLETION FUNCTIONS
 ****************************************************************************/


static void
change_one_item (void)
{
  if (compl_items == NULL)
    return;
  
  if (compl_items->count == 1){
    compl_string = xstrdup (compl_items->array[0]);
    rstring_delete (compl_items);
    compl_items = NULL;
  }
}


static int
is_prefix (const char *a, const char *b)
{
  while (*a){
    if (*a != *b)
      return 0;
    a++;
    b++;
  }
  return 1;
}


static int
select_fun (const struct dirent *entry)
{
  int   ret;
  char  c = *position;
  char *fname;

  *position = '\0';
  fname     = strrchr (buffer, '/');

  if (fname == NULL)
    fname = buffer;
  else
    fname++;

  if (*fname != '.' && *entry->d_name == '.'){
    *position = c;
    return 0;
  }
  
  ret       = is_prefix (fname, entry->d_name);
  *position = c;

  return ret;
}


static void
complete_prepare_funs (void)
{
  compl_items = exec_get_functions (buffer);
  change_one_item ();
}



static void
complete_prepare_files (void)
{
  char  c = *position;
  char *dir;
  char *ndir;
  char *seek;

  *position = '\0';
  seek      = strrchr (buffer, '/');
  *position = c;

  if (seek){
    dir = xmalloc (seek - buffer + 2);
    memcpy (dir, buffer, seek - buffer + 1);
    dir[seek - buffer + 1] = '\0';
    if (*dir == '~'){
      ndir = file_expand_tilde (dir);
      xfree (dir);
      dir  = ndir;
    }
  }
  else {
    dir = xstrdup (".");
  }

  compl_items = file_dir_items (dir, select_fun);
  xfree (dir);
  change_one_item ();
}


static void
complete_prepare_addrs (void)
{
  char  c = *position;
  char *seek;

  *position = '\0';
  seek      = strrchr (buffer, ',');
  *position = c;

  if (seek){
    do
      seek++;
    while (isspace (*seek));
  }
  else {
    seek = buffer;
  }
  
  compl_items = abook_strings (seek);
  change_one_item ();
}


static int
longest_common_prefix (rstring_t *vect)
{
  int i;
  int index = 0;

  if (vect->count == 0)
    return 0;
  
  if (vect->count == 1)
    return strlen (vect->array[0]);

  while (vect->array[0][index]){
    for (i = 1; i < vect->count; i++){
      if (vect->array[0][index] != vect->array[i][index])
        break;
    }
    if (i < vect->count)
      break;
    index++;
  }
  return index;
}


static void
replace (const char *str, int len)
{
  int   l;
  char  c;
  char *fname = NULL;
  char *addr  = NULL;
  
  if (len == 0)
    return;
  
  switch (mode){

    case COMPLETE_NONE: // no way !
      return;

    case COMPLETE_FILES:
      c         = *position;
      *position = '\0';
      fname     = strrchr (buffer, '/');
      *position = c;

      if (fname)
        fname++;
      else
        fname = buffer;

      l = strlen (position);
      if (len > position - fname && l > 0){
        memmove (fname + len, position, l + 1);
        memcpy (fname, str, len);
        position = fname + len;
      }
      else if (len > position - fname){
        memcpy (fname, str, len);
        position  = fname + len;
        *position = '\0';
      }
      break;

    case COMPLETE_FUNS:
      memcpy (buffer, str, len);
      buffer[len] = '\0';
      position    = buffer + len;
      break;

    case COMPLETE_ADDRS:
      c         = *position;
      *position = '\0';
      addr      = strrchr (buffer, ',');
      *position = c;

      if (addr)
        addr++;
      else
        addr = buffer;

      l = strlen (position);
      if (len > position - addr && l > 0){
        if (addr != buffer){
          addr[len] = ' ';
          addr++;
        }
        memmove (addr + len, position, l + 1);
        memcpy (fname, str, len);
        position = addr + len;
      }
      else if (len > position - addr){
        if (addr != buffer){
          *addr = ' ';
          addr++;
        }
        memcpy (addr, str, len);
        position  = addr + len;
        *position = '\0';
      }
      break;
  }
  wmove (cmd_echo_area, 0, prompt_len);
  wclrtoeol (cmd_echo_area);
  waddstr (cmd_echo_area, buffer);
  wmove (cmd_echo_area, 0, prompt_len + (position - buffer));
}



static void
complete (void)
{
  int len;
  
  switch (mode){

    case COMPLETE_NONE:
      return;

    case COMPLETE_FILES:
      complete_prepare_files ();
      break;

    case COMPLETE_FUNS:
      complete_prepare_funs ();
      break;

    case COMPLETE_ADDRS:
      complete_prepare_addrs ();
      break;
  }

  if (compl_string){
    replace (compl_string, strlen (compl_string));
    xfree (compl_string);
    compl_string = NULL;
  }
  else if (compl_items){
    fprintf (stderr, "%c", '\a');
    len = longest_common_prefix (compl_items);
    replace (compl_items->array[0], len);
    tab = 1;
  }
  else {
    fprintf (stderr, "%c", '\a');
  }
}



static void
choose_wrapper (void (*fun)(void))
{
  int y, x;

  getyx (cmd_echo_area, y, x);
  fun ();
  wmove (cmd_echo_area, y, x);
}



static void
leave_choose (void)
{
  if (tab){
    tab = 0;
    choose_wrapper (choose_close);
    cmd_refresh ();
  }

  if (compl_string)
    xfree (compl_string);

  if (compl_items)
    rstring_delete (compl_items);

  compl_string = NULL;
  compl_items  = NULL;
}


/****************************************************************************
 *    IMPLEMENTATION PRIVATE FUNCTIONS
 ****************************************************************************/

static void
action (int c)
{
  int index;
  
  switch (c){

    case 1:   /* ^A */
    case KEY_HOME:
      if (tab){
        choose_wrapper (choose_first);
      }
      else {
        first_char ();
      }
      break;
    
    case 2:   /* ^B */
    case KEY_LEFT:
      leave_choose ();
      prev_char ();
      break;

    case 4:      /* ^D  */
    case KEY_DC: /* del */
      leave_choose ();
      kill_char ();
      break;
    
    case 5:   /* ^E */
    case KEY_END:
      if (tab){
        choose_wrapper (choose_last);
      }
      else {
        last_char ();
      }
      break;
    
    case 6:   /* ^F */
    case KEY_RIGHT:
      leave_choose ();
      next_char ();
      break;

    case 7:   /* ^G  */
    case 27:  /* esc */
      if (tab){
        leave_choose ();
      }
      else {
        cancel = 1;
      }
      break;

    case 9:   /* TAB */
      if (mode == COMPLETE_NONE){
        leave_choose ();
        echo_char (' ');
      }
      else if (tab){
        int y, x;
        getyx (cmd_echo_area, y, x);
        choose_open (compl_items->count, compl_items->array);
        wmove (cmd_echo_area, y, x);
      }
      else {
        complete ();
      }
      break;
    
    case 11:  /* ^K */
      leave_choose ();
      kill_line ();
      break;

    case 14:  /* ^N */
    case KEY_DOWN:
      if (tab){
        choose_wrapper (choose_next);
      }
      break;

    case 16:  /* ^P */
    case KEY_UP:
      if (tab){
        choose_wrapper (choose_prev);
      }
      break;


    case KEY_BACKSPACE:
    case '\b':
    case 127:
      leave_choose ();
      backspace ();
      break;

    case '\r':
      index = choose_index ();
      replace (compl_items->array[index], strlen (compl_items->array[index]));
      leave_choose ();
      break;

    default:
      leave_choose ();
      echo_char (c & 0xff);
      break;
  }
}


static void
write_prompt (const char *prompt)
{
  int y;
  
  mvwaddstr (cmd_echo_area, 0, 0, prompt);
  getyx (cmd_echo_area, y, prompt_len);
  waddstr (cmd_echo_area, buffer);
}



static void
get_string (const char *prompt)
{
  int c;
  
  tab = 0;
  write_prompt (prompt);
  while (1){
    c = wgetch (cmd_echo_area);
    if (c == '\r' && tab == 0){
      break;
    }

    action (c);
    screen_doupdate ();
    if (cancel){
      break;
    }
  }
}



static void
init_buffer (const char *def)
{
  int len;
  
  wbkgd (cmd_echo_area, COLOR_PAIR (8));

  if (buffer == NULL){
    bufsize = COLS + 2;
    buffer  = xmalloc (bufsize);
  }

  if (def){
    len = strlen (def);
    memcpy (buffer, def, len + 1);
    position = buffer + len;
  }
  else {
    position  = buffer;
    *position = '\0';
  }
}



static void
clear_win (void)
{
  wmove (cmd_echo_area, 0, 0);
  wclrtoeol (cmd_echo_area);
}


/****************************************************************************
 *    INTERFACE FUNCTIONS
 ****************************************************************************/

char *
read_argument (const char *prompt, const char *def, complete_mode_t cmode)
{
  init_buffer (def);
  clear_win ();
  
  clock_stop ();
  curs_set (1);
  meta (cmd_echo_area, TRUE);
  
  mode = cmode;
  get_string (prompt);
  mode = COMPLETE_NONE;
  
  clear_win ();
  curs_set (0);
  clock_resume ();

  if (cancel){
    cancel = 0;
    return NULL;
  }

  return buffer;
}



void
read_free_resources (void)
{
  if (buffer){
    xfree (buffer);
    buffer  = NULL;
    bufsize = 0;
  }
}

/****************************************************************************
 *    INTERFACE CLASS BODIES
 ****************************************************************************/
/****************************************************************************
 *
 *    END MODULE read.c
 *
 ****************************************************************************/
