/* 
   elmo - ELectronic Mail Operator

   Copyright (C) 2002, 2003 rzyjontko

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  

   ----------------------------------------------------------------------

   This module implements RFC2821 and RFC821 (e)smtp client side protocol.

*/
/****************************************************************************
 *    IMPLEMENTATION HEADERS
 ****************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <errno.h>
#include <error.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <sys/stat.h>

#include "networking.h"
#include "smtp.h"
#include "error.h"
#include "file.h"
#include "address.h"
#include "raddress.h"
#include "gettext.h"
#include "xmalloc.h"
#include "mime.h"
#include "hmac-md5.h"
#include "str.h"

/****************************************************************************
 *    IMPLEMENTATION PRIVATE DEFINITIONS / ENUMERATIONS / SIMPLE TYPEDEFS
 ****************************************************************************/

#define RECV_BUFFER_SIZE    80
#define SEND_BUFFER_SIZE    75
#define HOSTNAME_MAX_LENGTH 255


#define SMTP_OK_211 211         /* system status or system help */
#define SMTP_OK_214 214         /* help message */
#define SMTP_OK_220 220         /* service ready */
#define SMTP_OK_221 221         /* service closing transmission channel */
#define SMTP_OK_235 235         /* authentication successful */
#define SMTP_OK_250 250         /* requested mail action okay */
#define SMTP_OK_251 251         /* user not local, will forward */
#define SMTP_OK_252 252         /* cannot vrfy user, but will accept message */

#define SMTP_INTER_354 354      /* start mail input */

#define SMTP_NEG_421 421        /* service not available, closing channel */
#define SMTP_NEG_450 450        /* action not taken: mailbox unavailable */
#define SMTP_NEG_451 451        /* action aborted: local error in processing */
#define SMTP_NEG_452 452        /* not taken: insufficient system storage */

#define SMTP_ERR_500 500        /* syntax error, command unrecognized */
#define SMTP_ERR_501 501        /* syntax error in parameters or arguments */
#define SMTP_ERR_502 502        /* command not implemented */
#define SMTP_ERR_503 503        /* bad sequence of commands */
#define SMTP_ERR_504 504        /* command parameter not implemented */
#define SMTP_ERR_550 550        /* action not taken: mailbox unavailable */
#define SMTP_ERR_551 551        /* user not local */
#define SMTP_ERR_552 552        /* action aborted: exceeded storage */
#define SMTP_ERR_553 553        /* not taken: mailbox name not allowed */
#define SMTP_ERR_554 554        /* transaction failed / no smtp service here */

#define NOT_SMTP_REPLY 666

/****************************************************************************
 *    IMPLEMENTATION PRIVATE CLASS PROTOTYPES / EXTERNAL CLASS REFERENCES
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE STRUCTURES / UTILITY CLASSES
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION REQUIRED EXTERNAL REFERENCES (AVOID)
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE DATA
 ****************************************************************************/

static char *challenge     = NULL;
static int   challenge_len = 0;

static char  hostname[HOSTNAME_MAX_LENGTH];

/****************************************************************************
 *    INTERFACE DATA
 ****************************************************************************/

int use_8bitmime_anyway = 1;

/****************************************************************************
 *    IMPLEMENTATION PRIVATE FUNCTION PROTOTYPES
 ****************************************************************************/

static char *my_helo_name (int fqdn);
static int   parse_ok_reply (const char *buf);
static int   parse_inter_reply (const char *buf);
static int   parse_neg_reply (const char *buf);
static int   parse_err_reply (const char *buf);
static int   parse_reply (const char *buf);
static int   parse_function (char *buf);
static int   negotiate_rcpt (char *name);
static int   is_complete (char *buf, size_t len);

/****************************************************************************
 *    IMPLEMENTATION PRIVATE FUNCTIONS
 ****************************************************************************/


static char *
my_helo_name (int fqdn){
  
  if (fqdn){
    gethostname (hostname, sizeof (hostname));
    return hostname;
  }
  else {
    return inet_ntoa (net_local_address);
  }
}



static int
parse_ok_reply (const char *buf){

  switch (*buf){
    case '1':
      if (*(buf + 1) == '1')
        return SMTP_OK_211;
      if (*(buf + 1) == '4')
        return SMTP_OK_214;
      return NOT_SMTP_REPLY;

    case '2':
      if (*(buf + 1) == '0')
        return SMTP_OK_220;
      if (*(buf + 1) == '1')
        return SMTP_OK_221;
      return NOT_SMTP_REPLY;

    case '3':
      if (*(buf + 1) == '5')
        return SMTP_OK_235;
      return NOT_SMTP_REPLY;

    case '5':
      if (*(buf + 1) == '0')
        return SMTP_OK_250;
      if (*(buf + 1) == '1')
        return SMTP_OK_251;
      if (*(buf + 1) == '2')
        return SMTP_OK_252;
      return NOT_SMTP_REPLY;

    default:
      return NOT_SMTP_REPLY;
  }
}



static int
parse_inter_reply (const char *buf){

  if (*buf == '5' && *(buf + 1) == '4')
    return SMTP_INTER_354;
  return NOT_SMTP_REPLY;
}



static int
parse_neg_reply (const char *buf){

  switch (*buf){
  case '2':
    if (*(buf + 1) == '1')
      return SMTP_NEG_421;
    return NOT_SMTP_REPLY;

  case '5':
    if (*(buf + 1) == '0')
      return SMTP_NEG_450;
    if (*(buf + 1) == '1')
      return SMTP_NEG_451;
    if (*(buf + 1) == '2')
      return SMTP_NEG_452;
    return NOT_SMTP_REPLY;

  default:
    return NOT_SMTP_REPLY;
  }
}



static int
parse_err_reply (const char *buf){

  switch (*buf){
  case '0':
    if (*(buf + 1) == '0')
      return SMTP_ERR_500;
    if (*(buf + 1) == '1')
      return SMTP_ERR_501;
    if (*(buf + 1) == '2')
      return SMTP_ERR_502;
    if (*(buf + 1) == '3')
      return SMTP_ERR_503;
    if (*(buf + 1) == '4')
      return SMTP_ERR_504;
    return NOT_SMTP_REPLY;

  case '5':
    if (*(buf + 1) == '0')
      return SMTP_ERR_550;
    if (*(buf + 1) == '1')
      return SMTP_ERR_551;
    if (*(buf + 1) == '2')
      return SMTP_ERR_552;
    if (*(buf + 1) == '3')
      return SMTP_ERR_553;
    if (*(buf + 1) == '4')
      return SMTP_ERR_554;
    return NOT_SMTP_REPLY;

  default:
    return NOT_SMTP_REPLY;
  }
}



static int
parse_reply (const char *buf){

  switch (*buf){
  case '2':
    return parse_ok_reply (buf + 1);

  case '3':
    return parse_inter_reply (buf + 1);

  case '4':
    return parse_neg_reply (buf + 1);

  case '5':
    return parse_err_reply (buf + 1);

  default:
    return NOT_SMTP_REPLY;
  }
}


static int
parse_auth (char *buf)
{
  mime_t *mime = mime_create (0);
  
  if (*buf != '3')
    return 1;

  challenge = strchr (buf, ' ');
  if (challenge == NULL)
    return 1;
  else
    challenge++;

  mime->offset_end = strlen (challenge);
  mime->encoding   = MENC_BASE64;

  challenge = mime_decode (mime, challenge, &challenge_len);

  mime_destroy (mime);
  return 0;
}


static int
parse_function (char *buf){
  int ret;

  ret = parse_reply (buf);

  switch (ret){

    case SMTP_OK_211: case SMTP_OK_214: case SMTP_OK_220: case SMTP_OK_221:
    case SMTP_OK_235: case SMTP_OK_250: case SMTP_OK_251: case SMTP_OK_252:
    case SMTP_INTER_354:
      return 0;

  case SMTP_NEG_421:
    error_ (0, _("service not available"));
    return 1;

  case SMTP_NEG_450:
    error_ (0, _("mailbox unavailable"));
    return 1;

  case SMTP_NEG_451:
    error_ (0, _("server error in processing"));
    return 1;

  case SMTP_NEG_452:
    error_ (0, _("server insufficient system storage"));
    return 1;

  case SMTP_ERR_500:
    error_ (0, _("syntax error"));
    return 1;

  case SMTP_ERR_501:
    error_ (0, _("syntax error in parameters / arguments"));
    return 1;

  case SMTP_ERR_502:
    error_ (0, _("command not implemented"));
    return 1;

  case SMTP_ERR_503:
    error_ (0, _("bad sequence of commands"));
    return 1;

  case SMTP_ERR_504:
    error_ (0, _("command parameter not implemented"));
    return 1;

  case SMTP_ERR_550:
    error_ (0, _("mailbox unavailable"));
    return 1;

  case SMTP_ERR_551:
    error_ (0, _("wrong server for this user"));
    return 1;

  case SMTP_ERR_552:
    error_ (0, _("server exceeded storage allocation"));
    return 1;

  case SMTP_ERR_553:
    error_ (0, _("invalid mailbox name"));
    return 1;

  case SMTP_ERR_554:
    error_ (0, _("connection failed"));
    return 1;

  case NOT_SMTP_REPLY:
    error_ (0, _("server does not conform to any RFC821 RFC2821"));
    return 1;

  default:
    error_ (0, _("internal error"));
    return 1;
  }
}



static int
negotiate_rcpt (char *name){
  
  return net_exchange_msg (parse_function, is_complete, "RCPT TO:<%s>\r\n",
                           name);
}




static int
is_complete (char *buf, size_t len){

  if (buf[len - 1] == '\n' && buf[len - 2] == '\r'){
    buf[len - 2] = '\0';
    return 1;
  }
  else {
    return 0;
  }
}



static char *
prepare_to_send (char *buf, int len, int *rlen)
{
  char *result    = xmalloc (2 * len);
  char *resultptr = result;
  char *seek      = buf;
  char *prev      = NULL;

  while (seek){
    prev = seek;
    seek  = strchr (seek, '\n');
    if (seek == NULL)
      break;
    if (*prev == '.'){
      resultptr[0] = '.';
      resultptr[1] = '.';
      resultptr   += 2;
      prev++;
    }
    memcpy (resultptr, prev, seek - prev);
    resultptr   += seek - prev;
    resultptr[0] = '\r';
    resultptr[1] = '\n';
    resultptr   += 2;
    seek++;
  }
  memcpy (resultptr, prev, len - (prev - buf));
  resultptr += len - (prev - buf);

  *rlen = resultptr - result;

  return result;
}



/****************************************************************************
 *    INTERFACE FUNCTIONS
 ****************************************************************************/


/*
  This only initiates smtp connection.  It connects to given host and
  waits for one of messages:
      - 220 service ready
      - 554 no smtp service here
  Send "QUIT" and close connection in second case.
*/
int
smtp_begin_connection (char *hostname, unsigned short port)
{

  /*
   * service opening takes too long
   */
  port = (port) ? port : 25;
  
  if (net_open_connection (hostname, port)){
    return 1;
  }

  return net_exchange_msg (parse_function, is_complete, NULL);
}



void
smtp_finish_connection (void)
{

  net_exchange_msg (parse_function, is_complete, "QUIT\r\n");
  net_close_connection ();
}



int
smtp_helo (int fqdn)
{

  return net_exchange_msg (parse_function, is_complete, "HELO %s\r\n",
                           my_helo_name (fqdn));
}



int
smtp_ehlo (int fqdn)
{
  int ret;

  ret = net_exchange_msg (parse_function, is_complete, "EHLO %s\r\n",
                          my_helo_name (fqdn));
  error_clear_force ();
  if (ret)
    return smtp_helo (fqdn);
  return 0;
}



int
smtp_auth (char *user, char *pass)
{
  int    ret;
  int    i;
  int    user_len = strlen (user);
  int    pass_len = strlen (pass);
  char   digest[16];
  str_t *str;
  char  *answer;
  char  *encoded;
  int    enc_len;
  
  if (net_exchange_msg (parse_auth, is_complete, "AUTH CRAM-MD5\r\n")){
    error_ (0, _("cram-md5 authentication method not supported by server"));
    return 1;
  }

  hmac_md5 (challenge, challenge_len, pass, pass_len, digest);

  str = str_create_size (user_len + 1 + 32 + 1);
  str_sprintf (str, "%s ", user);
  for (i = 0; i < 16; i++){
    str_sprintf (str, "%02x", (unsigned char) digest[i]);
  }

  answer  = str_finished (str);
  encoded = mime_encode (answer, strlen (answer), &enc_len, MENC_BASE64);
  encoded[enc_len - 1] = '\0';

  ret = net_exchange_msg (parse_function, is_complete, "%s\r\n", encoded);

  xfree (answer);
  xfree (encoded);
  xfree (challenge);

  challenge     = NULL;
  challenge_len = 0;
  
  return ret;
}



int
smtp_mail_from (const char *email)
{

  return net_exchange_msg (parse_function, is_complete, "MAIL FROM:<%s>\r\n",
                           email);
}



int
smtp_send_file (const char *fname)
{
  char *buffer;
  char *send_buf;
  int   size;
  int   newsize;
  FILE *fp = fopen (fname, "r");

  if (fp == NULL){
    error_ (errno, "%s", fname);
    return 1;
  }
  
  if (file_whole (fp, &buffer, &size)){
    fclose (fp);
    return 1;
  }
  fclose (fp);

  send_buf = prepare_to_send (buffer, size, &newsize);
  file_free_buffer (buffer, size);

  size = net_send (send_buf, newsize);
  xfree (send_buf);
  
  return size != newsize;
}



int
smtp_send_mail (const char *file, raddress_t *receivers)
{
  address_t **rcpt;
  int      success_count = 0;
  int      ret;

  if (receivers == NULL || receivers->count == 0){
    return 0;
  }
  
  for (rcpt = receivers->array; *rcpt; rcpt++){
    if (negotiate_rcpt ((*rcpt)->email) == 0)
      success_count++;
    else
      error_ (0, _("address <%s> was rejected"), (*rcpt)->email);
  }

  if (!success_count)
    return 1;
   
  if (net_exchange_msg (parse_function, is_complete, "DATA\r\n"))
    return 1;

  ret = smtp_send_file (file);

  net_exchange_msg (parse_function, is_complete, "\r\n.\r\n");

  return ret;
}



/****************************************************************************
 *    INTERFACE CLASS BODIES
 ****************************************************************************/
/****************************************************************************
 *
 *    END MODULE smtp.c
 *
 ****************************************************************************/
