/*
   elmo - ELectronic Mail Operator

   Copyright (C) 2002 rzyjontko

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  

   ----------------------------------------------------------------------

   xmalloc - memory allocation with out of memory checking
             and memory leaks debugging


*/

/****************************************************************************
 *    IMPLEMENTATION HEADERS
 ****************************************************************************/

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>

#ifdef XMALLOC_DEBUG
# include <execinfo.h>
# include <malloc.h>
#endif


#include "error.h"

/****************************************************************************
 *    IMPLEMENTATION PRIVATE DEFINITIONS / ENUMERATIONS / SIMPLE TYPEDEFS
 ****************************************************************************/

#define FRAME_NO 2

/****************************************************************************
 *    IMPLEMENTATION PRIVATE CLASS PROTOTYPES / EXTERNAL CLASS REFERENCES
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE STRUCTURES / UTILITY CLASSES
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION REQUIRED EXTERNAL REFERENCES (AVOID)
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE DATA
 ****************************************************************************/

#ifdef XMALLOC_DEBUG

static FILE *fp = NULL;

#endif

/****************************************************************************
 *    INTERFACE DATA
 ****************************************************************************/
/****************************************************************************
 *    IMPLEMENTATION PRIVATE FUNCTION PROTOTYPES
 ****************************************************************************/

#ifdef XMALLOC_DEBUG

static void my_init_hook (void);

/* Override initializing hook from the C library. */
void (*__malloc_initialize_hook) (void) = my_init_hook;

#endif

/****************************************************************************
 *    IMPLEMENTATION PRIVATE FUNCTIONS
 ****************************************************************************/

#ifdef XMALLOC_DEBUG

static void
my_init_hook (void)
{
  /* Flawfinder: ignore */
  char *xmalloc_trace = getenv ("XMALLOC_TRACE");

  if (xmalloc_trace)
    fp = fopen (xmalloc_trace, "w");
  
  if (fp)
    fprintf (fp, "= Start\n");
}

#endif

/****************************************************************************
 *    INTERFACE FUNCTIONS
 ****************************************************************************/


void *
xmalloc (size_t n)
{

#ifdef XMALLOC_DEBUG
  void    *array[FRAME_NO];
  size_t   s;
  char   **strings;
  char    *string;
  char    *tmp;
#endif

  void *p;

  if (n == 0)
    return NULL;

  p = malloc (n);


#ifdef XMALLOC_DEBUG
  if (fp){
    
    s       = backtrace (array, FRAME_NO);
    strings = backtrace_symbols (array, s);

    string = strings[FRAME_NO - 1];
    tmp    = strchr (string, '(');

    if (tmp){
      *tmp   = '\0';
      fprintf (fp, "@ %s:", string);
      *tmp   = '(';
      string = strchr (tmp, ' ');

      if (string){
        *string = '\0';
        fprintf (fp, "%s%s + %p 0x%x\n", tmp, string + 1, p, n);
        *string = ' ';
      }
    }
    else {
      tmp        = strchr (string, '[');
      *(tmp - 1) = '\0';
      fprintf (fp, "@ %s:", string);
      *(tmp - 1) = '[';
      fprintf (fp, "%s + %p 0x%x\n", tmp, p, n);
    }
    free (strings);
  }
#endif
  
  if (p == NULL)
    error_critical (EXIT_FAILURE, errno, "malloc");
  return p;
}




void *
xcalloc (n, s)
     size_t n, s;
{

#ifdef XMALLOC_DEBUG
  void    *array[FRAME_NO];
  size_t   cnt;
  char   **strings;
  char    *string;
  char    *tmp;
#endif

  void *p;

  if (n == 0 || s == 0)
    return NULL;

  p = calloc (n, s);

#ifdef XMALLOC_DEBUG
  if (fp){
    
    cnt     = backtrace (array, FRAME_NO);
    strings = backtrace_symbols (array, cnt);

    string = strings[FRAME_NO - 1];
    tmp    = strchr (string, '(');

    if (tmp){
      *tmp   = '\0';
      fprintf (fp, "@ %s:", string);
      *tmp   = '(';
      string = strchr (tmp, ' ');

      if (string){
        *string = '\0';
        fprintf (fp, "%s%s + %p 0x%x\n", tmp, string + 1, p, n * s);
        *string = ' ';
      }
    }
    else {
      tmp        = strchr (string, '[');
      *(tmp - 1) = '\0';
      fprintf (fp, "@ %s:", string);
      *(tmp - 1) = '[';
      fprintf (fp, "%s + %p 0x%x\n", tmp, p, n * s);
    }
    free (strings);
  }
#endif
  
  if (p == NULL)
    error_critical (EXIT_FAILURE, errno, "calloc");
  return p;
}




void *
xrealloc (p, n)
     void *p;
     size_t n;
{

#ifdef XMALLOC_DEBUG
  void    *array[FRAME_NO];
  size_t   s;
  char   **strings;
  char    *string;
  char    *tmp;
#endif

  void *result;

  if (p == 0)
    return xmalloc (n);
  if (n == 0)
    return NULL;

  result = realloc (p, n);

#ifdef XMALLOC_DEBUG
  if (fp){
    
    s       = backtrace (array, FRAME_NO);
    strings = backtrace_symbols (array, s);

    string = strings[FRAME_NO - 1];
    tmp    = strchr (string, '(');

    if (tmp){
      *tmp   = '\0';
      fprintf (fp, "@ %s:", string);
      *tmp   = '(';
      string = strchr (tmp, ' ');

      if (string){
        *string = '\0';
        fprintf (fp, "%s%s < %p\n", tmp, string + 1, p);
        *string = ' ';
      }
    }
    else {
      tmp        = strchr (string, '[');
      *(tmp - 1) = '\0';
      fprintf (fp, "@ %s:", string);
      *(tmp - 1) = '[';
      fprintf (fp, "%s < %p\n", tmp, p);
    }


    string = strings[1];
    tmp    = strchr (string, '(');

    if (tmp){
      *tmp   = '\0';
      fprintf (fp, "@ %s:", string);
      *tmp   = '(';
      string = strchr (tmp, ' ');

      if (string){
        *string = '\0';
        fprintf (fp, "%s%s > %p 0x%x\n", tmp, string + 1, result, n);
        *string = ' ';
      }
    }
    else {
      tmp        = strchr (string, '[');
      *tmp = '\0';
      fprintf (fp, "@ %s:", string);
      *tmp = '[';
      fprintf (fp, "%s > %p 0x%x\n", tmp + 1, result, n);
    }
    free (strings);
  }
#endif  
  
  if (result == NULL)
    error_critical (EXIT_FAILURE, errno, "realloc");
  return result;
}





char *
xstrdup (str)
  const char *str;
{

#ifdef XMALLOC_DEBUG
  void    *array[FRAME_NO];
  size_t   s;
  char   **strings;
  char    *string;
  char    *tmp;
#endif

  char *p;

  if (str == NULL)
    return NULL;

  p = strdup (str);

#ifdef XMALLOC_DEBUG
  if (fp){
    
    s       = backtrace (array, FRAME_NO);
    strings = backtrace_symbols (array, s);

    string = strings[FRAME_NO - 1];
    tmp    = strchr (string, '(');

    if (tmp){
      *tmp   = '\0';
      fprintf (fp, "@ %s:", string);
      *tmp   = '(';
      string = strchr (tmp, ' ');

      if (string){
        *string = '\0';
        fprintf (fp, "%s%s + %p 0x%x\n", tmp, string + 1, p, strlen (str));
        *string = ' ';
      }
    }
    else {
      tmp        = strchr (string, '[');
      *(tmp - 1) = '\0';
      fprintf (fp, "@ %s:", string);
      *(tmp - 1) = '[';
      fprintf (fp, "%s + %p 0x%x\n", tmp, p, strlen (str));
    }
    free (strings);
  }
#endif  

  if (p == NULL)
    error_critical (EXIT_FAILURE, errno, "strdup");
  return p;
}



void
xfree (void *ptr){

#ifdef XMALLOC_DEBUG
  void    *array[FRAME_NO];
  size_t   s;
  char   **strings;
  char    *string;
  char    *tmp;
#endif

  free (ptr);
  
#ifdef XMALLOC_DEBUG
  if (fp){
    s       = backtrace (array, FRAME_NO);
    strings = backtrace_symbols (array, s);

    string = strings[FRAME_NO - 1];
    tmp    = strchr (string, '(');

    if (tmp){
      *tmp   = '\0';
      fprintf (fp, "@ %s:", string);
      *tmp   = '(';
      string = strchr (tmp, ' ');

      if (string){
        *string = '\0';
        fprintf (fp, "%s%s - %p\n", tmp, string + 1, ptr);
        *string = ' ';
      }
    }
    else {
      tmp        = strchr (string, '[');
      *(tmp - 1) = '\0';
      fprintf (fp, "@ %s:", string);
      *(tmp - 1) = '[';
      fprintf (fp, "%s - %p\n", tmp, ptr);
    }
  
    free (strings);
  }  
#endif
}



/****************************************************************************
 *    INTERFACE CLASS BODIES
 ****************************************************************************/
/****************************************************************************
 *
 *    END MODULE xmalloc.c
 *
 ****************************************************************************/
